# -*- coding: utf-8 -*-

# =============================================================================
# Global Sensitivity Analysis (GSA) functions and class for the Delta
# Moment-Independent measure based on Monte Carlo simulation LCA results.
# see: https://salib.readthedocs.io/en/latest/api.html#delta-moment-independent-measure
# =============================================================================
import os
import traceback
from time import time
from logging import getLogger

import bw2calc as bc
import numpy as np
import pandas as pd
from SALib.analyze import delta

from activity_browser.mod import bw2data as bd

from ..settings import ab_settings
from .montecarlo import MonteCarloLCA, perform_MonteCarlo_LCA

try:
    # attempt bw25 import
    from bw_graph_tools.graph_traversal import NewNodeEachVisitGraphTraversal
except ImportError:
    # standard import on failure
    from bw2calc import GraphTraversal


log = getLogger(__name__)


def get_lca(fu, method):
    """Calculates a non-stochastic LCA and returns a the LCA object."""
    lca = bc.LCA(fu, method=method)
    lca.lci()
    lca.lcia()
    log.info(f"Non-stochastic LCA score: {lca.score}")

    # add reverse dictionaries
    lca.activity_dict_rev, lca.product_dict_rev, lca.biosphere_dict_rev = (
        lca.reverse_dict()
    )

    return lca


def filter_technosphere_exchanges(fu, method, cutoff=0.05, max_calc=1000):
    """Use brightway's GraphTraversal to identify the relevant
    technosphere exchanges in a non-stochastic LCA."""
    start = time()
    res = NewNodeEachVisitGraphTraversal.calculate(fu, method, cutoff=cutoff, max_calc=int(max_calc))

    # get all edges
    technosphere_exchange_indices = []
    for e in res["edges"]:
        if e["to"] != -1:  # filter out head introduced in graph traversal
            technosphere_exchange_indices.append((e["from"], e["to"]))
    log.info(
        "TECHNOSPHERE {} filtering resulted in {} of {} exchanges and took {} iterations in {} seconds.".format(
            res["lca"].technosphere_matrix.shape,
            len(technosphere_exchange_indices),
            res["lca"].technosphere_matrix.getnnz(),
            res["counter"],
            np.round(time() - start, 2),
        )
    )
    return technosphere_exchange_indices


def filter_biosphere_exchanges(lca, cutoff=0.005):
    """Reduce biosphere exchanges to those that matter for a given impact
    category in a non-stochastic LCA."""
    start = time()

    inv = lca.characterized_inventory
    finv = inv.multiply(abs(inv) > abs(lca.score / (1 / cutoff)))
    biosphere_exchange_indices = list(zip(*finv.nonzero()))
    explained_fraction = finv.sum() / lca.score
    log.info(
        "BIOSPHERE {} filtering resulted in {} of {} exchanges ({}% of total impact) and took {} seconds.".format(
            inv.shape,
            finv.nnz,
            inv.nnz,
            np.round(explained_fraction * 100, 2),
            np.round(time() - start, 2),
        )
    )
    return biosphere_exchange_indices


def get_exchanges(lca, indices, biosphere=False, only_uncertain=True):
    """Get actual exchange objects from indices.
    By default get only exchanges that have uncertainties.

    Returns
    -------
    exchanges : list
        List of exchange objects
    indices : list of tuples
        List of indices
    """
    exchanges = list()
    for i in indices:
        if biosphere:
            from_act = bd.get_activity(lca.biosphere_dict_rev[i[0]])
        else:  # technosphere
            from_act = bd.get_activity(lca.activity_dict_rev[i[0]])
        to_act = bd.get_activity(lca.activity_dict_rev[i[1]])

        for exc in to_act.exchanges():
            if exc.input == from_act.key:
                exchanges.append(exc)
                # continue  # if there was always only one max exchange between two activities

    # in theory there should be as many exchanges as indices, but since
    # multiple exchanges are possible between two activities, the number of
    # exchanges must be at least equal or higher to the number of indices
    if len(exchanges) < len(
        indices
    ):  # must have at least as many exchanges as indices (assu)
        raise ValueError(
            "Error: mismatch between indices provided ({}) and Exchanges received ({}).".format(
                len(indices), len(exchanges)
            )
        )

    # by default drop exchanges and indices if the have no uncertainties
    if only_uncertain:
        exchanges, indices = drop_no_uncertainty_exchanges(exchanges, indices)

    return exchanges, indices


def drop_no_uncertainty_exchanges(excs, indices):
    excs_no = list()
    indices_no = list()
    for exc, ind in zip(excs, indices):
        if exc.get("uncertainty type") and exc.get("uncertainty type") >= 1:
            excs_no.append(exc)
            indices_no.append(ind)
    log.info(
        "Dropping {} exchanges of {} with no uncertainty. {} remaining.".format(
            len(excs) - len(excs_no), len(excs), len(excs_no)
        )
    )
    return excs_no, indices_no


def get_exchanges_dataframe(exchanges, indices, biosphere=False):
    """Returns a Dataframe from the exchange data and a bit of additional information."""

    for exc, i in zip(exchanges, indices):
        from_act = bd.get_activity(exc.get("input"))
        to_act = bd.get_activity(exc.get("output"))

        exc.update(
            {
                "index": i,
                "from name": from_act.get("name", np.nan),
                "from location": from_act.get("location", np.nan),
                "to name": to_act.get("name", np.nan),
                "to location": to_act.get("location", np.nan),
            }
        )

        # GSA name (needs to yield unique labels!)
        if biosphere:
            exc.update(
                {
                    "GSA name": "B: {} // {} ({}) [{}]".format(
                        from_act.get("name", ""),
                        to_act.get("name", ""),
                        to_act.get("reference product", ""),
                        to_act.get("location", ""),
                    )
                }
            )
        else:
            exc.update(
                {
                    "GSA name": "T: {} FROM {} [{}] TO {} ({}) [{}]".format(
                        from_act.get("reference product", ""),
                        from_act.get("name", ""),
                        from_act.get("location", ""),
                        to_act.get("name", ""),
                        to_act.get("reference product", ""),
                        to_act.get("location", ""),
                    )
                }
            )

    return pd.DataFrame(exchanges)


def get_CF_dataframe(lca, only_uncertain_CFs=True):
    """Returns a dataframe with the metadata for the characterization factors
    (in the biosphere matrix). Filters non-stochastic CFs if desired (default)."""
    data = dict()
    for params_index, row in enumerate(lca.cf_params):
        if only_uncertain_CFs and row["uncertainty_type"] <= 1:
            continue
        cf_index = row["row"]
        bio_act = bd.get_activity(lca.biosphere_dict_rev[cf_index])

        data.update({params_index: bio_act.as_dict()})

        for name in row.dtype.names:
            data[params_index][name] = row[name]

        data[params_index]["index"] = cf_index
        data[params_index]["GSA name"] = (
            "CF: " + bio_act["name"] + str(bio_act["categories"])
        )

    log.info(
        "CHARACTERIZATION FACTORS filtering resulted in including {} of {} characteriation factors.".format(
            len(data),
            len(lca.cf_params),
        )
    )
    df = pd.DataFrame(data).T
    df.rename(columns={"uncertainty_type": "uncertainty type"}, inplace=True)
    return df


def get_parameters_DF(mc):
    """Product to make parameters dataframe"""
    if bool(mc.parameter_data):  # returns False if dict is empty
        dfp = pd.DataFrame(mc.parameter_data).T
        dfp["GSA name"] = "P: " + dfp["name"]
        log.info(f"PARAMETERS: {len(dfp)}")
        return dfp
    else:
        log.info("PARAMETERS: None included.")
        return pd.DataFrame()  # return emtpy df


def get_exchange_values(matrix, indices):
    """Get technosphere exchanges values from a list of exchanges
    (row and column information)"""
    return [matrix[i] for i in indices]


def get_X(matrix_list, indices):
    """Get the input data to the GSA, i.e. A and B matrix values for each
    model run."""
    X = np.zeros((len(matrix_list), len(indices)))
    for row, M in enumerate(matrix_list):
        X[row, :] = get_exchange_values(M, indices)
    return X


def get_X_CF(mc, dfcf, method):
    """Get the characterization factors used for each model run. Only those CFs
    that are in the dfcf dataframe will be returned (i.e. by default only the
    CFs that have uncertainties."""
    # get all CF inputs
    CF_data = np.array(mc.CF_dict[method])  # has the same shape as the Xa and Xb below

    # reduce this to uncertain CFs only (if this was done for the dfcf)
    params_indices = dfcf.index.values

    # if params_indices:
    return CF_data[:, params_indices]


def get_X_P(dfp):
    """Get the parameter values for each model run"""
    lists = [d for d in dfp["values"]]
    return list(zip(*lists))


def get_problem(X, names):
    return {
        "num_vars": X.shape[1],
        "names": names,
        "bounds": list(zip(*(np.amin(X, axis=0), np.amax(X, axis=0)))),
    }


class GlobalSensitivityAnalysis(object):
    """Class for Global Sensitivity Analysis.
    For now Delta Moment Independent Measure based on:
    https://salib.readthedocs.io/en/latest/api.html#delta-moment-independent-measure
    Builds on top of Monte Carlo Simulation results.
    """

    def __init__(self, mc):
        self.update_mc(mc)
        self.act_number = int()
        self.method_number = int()
        self.cutoff_technosphere = float()
        self.cutoff_biosphere = float()

    def update_mc(self, mc):
        "Update the Monte Carlo Simulation object (and results)."
        try:
            assert isinstance(mc, MonteCarloLCA)
            self.mc = mc
        except AssertionError:
            raise AssertionError(
                "mc should be an instance of MonteCarloLCA, but instead it is a {}.".format(
                    type(mc)
                )
            )

    def perform_GSA(
        self,
        act_number=0,
        method_number=0,
        cutoff_technosphere=0.01,
        cutoff_biosphere=0.01,
    ):
        """Perform GSA for specific reference flow and impact category."""
        start = time()

        # set FU and method
        try:
            self.act_number = act_number
            self.method_number = method_number
            self.cutoff_technosphere = cutoff_technosphere
            self.cutoff_biosphere = cutoff_biosphere

            self.fu = self.mc.cs["inv"][act_number]
            self.activity = bd.get_activity(self.mc.rev_activity_index[act_number])
            self.method = self.mc.cs["ia"][method_number]

        except Exception as e:
            traceback.print_exc()
            # todo: QMessageBox.warning(self, 'Could not perform Delta analysis', str(e))
            log.error("Initializing the GSA failed.")
            return None

        log.info(
            f"-- GSA --\n Project: {bd.projects.current} CS: {self.mc.cs_name} "
            f"Activity: {self.activity} Method: {self.method}",
        )

        # get non-stochastic LCA object with reverse dictionaries
        self.lca = get_lca(self.fu, self.method)

        # =============================================================================
        #   Filter exchanges and get metadata DataFrames
        # =============================================================================
        dfs = []
        # technosphere
        if self.mc.include_technosphere:
            self.t_indices = filter_technosphere_exchanges(
                self.fu, self.method, cutoff=cutoff_technosphere, max_calc=1e4
            )
            self.t_exchanges, self.t_indices = get_exchanges(self.lca, self.t_indices)
            self.dft = get_exchanges_dataframe(self.t_exchanges, self.t_indices)
            if not self.dft.empty:
                dfs.append(self.dft)

        # biosphere
        if self.mc.include_biosphere:
            self.b_indices = filter_biosphere_exchanges(
                self.lca, cutoff=cutoff_biosphere
            )
            self.b_exchanges, self.b_indices = get_exchanges(
                self.lca, self.b_indices, biosphere=True
            )
            self.dfb = get_exchanges_dataframe(
                self.b_exchanges, self.b_indices, biosphere=True
            )
            if not self.dfb.empty:
                dfs.append(self.dfb)

        # characterization factors
        if self.mc.include_cfs:
            self.dfcf = get_CF_dataframe(
                self.lca, only_uncertain_CFs=True
            )  # None if no stochastic CFs
            if not self.dfcf.empty:
                dfs.append(self.dfcf)

        # parameters
        # todo: if parameters, include df, but remove exchanges from T and B (skipped for now)
        self.dfp = get_parameters_DF(self.mc)  # Empty df if no parameters
        if not self.dfp.empty:
            dfs.append(self.dfp)

        # Join dataframes to get metadata
        self.metadata = pd.concat(dfs, axis=0, ignore_index=True, sort=False)
        self.metadata.set_index("GSA name", inplace=True)

        # =============================================================================
        #     GSA
        # =============================================================================

        # Get X (Technosphere, Biosphere and CF values)
        X_list = list()
        if self.mc.include_technosphere and self.t_indices:
            self.Xa = get_X(self.mc.A_matrices, self.t_indices)
            X_list.append(self.Xa)
        if self.mc.include_biosphere and self.b_indices:
            self.Xb = get_X(self.mc.B_matrices, self.b_indices)
            X_list.append(self.Xb)
        if self.mc.include_cfs and not self.dfcf.empty:
            self.Xc = get_X_CF(self.mc, self.dfcf, self.method)
            X_list.append(self.Xc)
        if self.mc.include_parameters and not self.dfp.empty:
            self.Xp = get_X_P(self.dfp)
            X_list.append(self.Xp)

        self.X = np.concatenate(X_list, axis=1)
        # print('X', self.X.shape)

        # Get Y (LCA scores)
        self.Y = self.mc.get_results_dataframe(act_key=self.activity.key)[
            self.method
        ].to_numpy()

        # log-transformation. This makes it more robust for very uneven distributions of LCA results (e.g. toxicity related impacts).
        # Can only be applied if all Monte-Carlo LCA scores are either positive or negative.
        # Should not be used when LCA scores overlap zero (sometimes positive and sometimes negative)
        # if np.all(self.Y > 0) if self.Y[0] > 0 else np.all(self.Y < 0):  # check if all LCA scores are of the same sign
        #     self.Y = np.log(np.abs(self.Y))  # this makes it more robust for very uneven distributions of LCA results
        if np.all(self.Y > 0):  # all positive numbers
            self.Y = np.log(np.abs(self.Y))
            log.info("All positive LCA scores. Log-transformation performed.")
        elif np.all(self.Y < 0):  # all negative numbers
            self.Y = -np.log(np.abs(self.Y))
            log.info("All negative LCA scores. Log-transformation performed.")
        else:  # mixed positive and negative numbers
            log.warning(
                "Log-transformation cannot be applied as LCA scores overlap zero."
            )

        # print('Filtering took {} seconds'.format(np.round(time() - start, 2)))

        # define problem
        self.names = self.metadata.index  # ['GSA name']
        # print('Names:', len(self.names))
        self.problem = get_problem(self.X, self.names)

        # perform delta analysis
        time_delta = time()
        self.Si = delta.analyze(self.problem, self.X, self.Y, print_to_console=False)
        log.info(
            "Delta analysis took {} seconds".format(
                np.round(time() - time_delta, 2),
            )
        )

        # put GSA results in to dataframe
        self.dfgsa = pd.DataFrame(self.Si, index=self.names).sort_values(
            by="delta", ascending=False
        )
        self.dfgsa.index.names = ["GSA name"]

        # join with metadata
        self.df_final = self.dfgsa.join(self.metadata, on="GSA name")
        self.df_final.reset_index(inplace=True)
        self.df_final["pedigree"] = [str(x) for x in self.df_final["pedigree"]]

        log.info("GSA took {} seconds".format(np.round(time() - start, 2)))

    def get_save_name(self):
        save_name = (
            self.mc.cs_name
            + "_"
            + str(self.mc.iterations)
            + "_"
            + self.activity["name"]
            + "_"
            + str(self.method)
            + ".xlsx"
        )
        save_name = save_name.replace(",", "").replace("'", "").replace("/", "")
        return save_name

    def export_GSA_output(self):
        save_name = "gsa_output_" + self.get_save_name()
        self.df_final.to_excel(os.path.join(ab_settings.data_dir, save_name))

    def export_GSA_input(self):
        """Export the input data to the GSA with a human readible index"""
        X_with_index = pd.DataFrame(self.X.T, index=self.metadata.index)
        save_name = "gsa_input_" + self.get_save_name()
        X_with_index.to_excel(os.path.join(ab_settings.data_dir, save_name))


if __name__ == "__main__":
    mc = perform_MonteCarlo_LCA(project="ei34", cs_name="kraft paper", iterations=20)
    g = GlobalSensitivityAnalysis(mc)
    g.perform_GSA(
        act_number=0, method_number=1, cutoff_technosphere=0.01, cutoff_biosphere=0.01
    )
    g.export()



import re
import os
from logging import getLogger
from copy import deepcopy

import requests

import ecoinvent_interface as ei
import bw2data as bd
import bw2io as bi

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Signal, SignalInstance

from activity_browser import application, signals
from activity_browser.ui import widgets, icons, threading
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils.io.ecoinvent_importer import Ecoinvent7zImporter
from activity_browser.bwutils.io.ecoinvent_lcia_importer import EcoinventLCIAImporter
from activity_browser.mod.bw2io.migrations import ab_create_core_migrations

log = getLogger(__name__)


class DatabaseImportFromEcoinvent(ABAction):
    """
    Launches the `EiWizard` dialog for importing a database from ecoinvent.

    This method creates an instance of the `EiWizard` class, passing the
    main application window as its parent, and executes the wizard dialog.

    Raises:
        Any exceptions encountered during the execution of the wizard
        are handled by the `exception_dialogs` decorator.
    """

    icon = icons.qicons.import_db
    text = "Import database from ecoinvent"
    tool_tip = "Import database from ecoinvent"

    @staticmethod
    @exception_dialogs
    def run():
        setup = EiWizard(application.main_window)
        setup.setWindowTitle("Import from ecoinvent")
        setup.exec_()


class EiWizard(widgets.ABWizard):
    """Wizard for importing database from ecoinvent"""

    class RemoteOrLocalPage(widgets.ABWizardPage):
        """Wizard page to choose between remote or local ecoinvent release"""
        title = "Import from ecoinvent"
        subtitle = "Choose whether to import from a remote or local ecoinvent release."

        def __init__(self, parent=None):
            super().__init__(parent)

            self.remote_button = QtWidgets.QRadioButton("Remote")
            self.local_button = QtWidgets.QRadioButton("Local")
            self.remote_button.setChecked(True)

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.remote_button)
            layout.addWidget(self.local_button)
            self.setLayout(layout)

        def nextPage(self):
            """Determine the next page based on the user's selection"""
            if self.local_button.isChecked():
                return EiWizard.LocalSelectPage
            else:
                return EiWizard.LoginPage

    class LocalSelectPage(widgets.ABWizardPage):
        """Wizard page to select a local ecoinvent .7z file"""
        title = "Import from ecoinvent"
        subtitle = "Select local ecoinvent .7z."

        def __init__(self, parent=None):
            super().__init__(parent)

            self.file_selector = widgets.ABFileSelector(filter="*.7z")
            self.file_selector.textChanged.connect(lambda: self.completeChanged.emit())

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.file_selector)
            self.setLayout(layout)

        def finalize(self, context: dict):
            """Store the selected file path in the context"""
            path = self.file_selector.text()
            context["ei_filepath"] = path
            # Try to extract version and model from the filename
            if version:= re.search(r'ecoinvent ([\d.]+)', path):
                context["version"] = version.group(1)
            if model:= re.search(r'ecoinvent [\d.]+_([^_]+)', path):
                context["model"] = model.group(1)

        def isComplete(self):
            """Check if a file has been selected"""
            return bool(self.file_selector.text())

        def nextPage(self):
            """Proceed to the BiosphereSetupPage"""
            return EiWizard.BiosphereSetupPage

    class LoginPage(widgets.ABWizardPage):
        """Wizard page to login with ecoinvent credentials"""
        title = "Login"
        subtitle = "Login with your ecoinvent credentials to authorize the download"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.release = None

            self.username = QtWidgets.QLineEdit()
            self.username.setPlaceholderText('ecoinvent username')

            self.password = QtWidgets.QLineEdit()
            self.password.setPlaceholderText('ecoinvent password')
            self.password.setEchoMode(QtWidgets.QLineEdit.Password)

            self.message = QtWidgets.QLabel()

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.username)
            layout.addWidget(self.password)
            layout.addWidget(self.message)

            self.setLayout(layout)

        def initializePage(self, context: dict):
            """Initialize the page with stored username and password"""
            settings = ei.Settings()
            self.username.setText(settings.username)
            self.password.setText(settings.password)

        def validatePage(self):
            """Validate the login credentials by attempting to list ecoinvent versions"""
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            try:
                settings = ei.Settings(username=self.username.text(), password=self.password.text())
                self.release = ei.EcoinventRelease(settings)
                self.release.list_versions()
            except requests.exceptions.HTTPError as e:
                QtWidgets.QApplication.restoreOverrideCursor()
                if e.response.status_code == 401:
                    self.message.setText("Invalid username and/or password, please try again.")
                    return False
                else:
                    self.message.setText("Unknown connection error, try again later.")
                    raise e
            ei.permanent_setting("username", self.username.text())
            ei.permanent_setting("password", self.password.text())
            QtWidgets.QApplication.restoreOverrideCursor()
            return True

        def finalize(self, context: dict):
            """Store the release object in the context"""
            context["release"] = self.release

        def nextPage(self):
            """Proceed to the EcoinventVersionPage"""
            return EiWizard.EcoinventVersionPage

    class EcoinventVersionPage(widgets.ABWizardPage):
        """Wizard page to choose ecoinvent version and system model"""
        title = "Choose version"
        subtitle = "Choose ecoinvent version and system model"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.versions = QtWidgets.QComboBox()
            self.models = QtWidgets.QComboBox()

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.versions)
            layout.addWidget(self.models)

            self.setLayout(layout)

        def initializePage(self, context: dict):
            """Initialize the page with available versions and models"""
            self.release = context["release"]
            self.versions.currentTextChanged.connect(self.collect_models)
            self.versions.addItems(self.release.list_versions())

        def finalize(self, context: dict):
            """Store the selected version and model in the context"""
            context["version"] = self.versions.currentText()
            context["model"] = self.models.currentText()

        def nextPage(self):
            """Proceed to the EcoinventDownloadPage"""
            return EiWizard.EcoinventDownloadPage

        def collect_models(self, version: str):
            """Collect and display system models for the selected version"""
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            self.models.clear()
            self.models.addItems(self.release.list_system_models(version))
            QtWidgets.QApplication.restoreOverrideCursor()

    class EcoinventDownloadPage(widgets.ABThreadedWizardPage):
        """Wizard page to download the selected ecoinvent release"""
        title = "Download ecoinvent"
        subtitle = "Downloading the selected ecoinvent release"

        class Thread(threading.ABThread):
            """Thread to handle the download process"""
            download_ready: SignalInstance = Signal(str)

            def run_safely(self, release: ei.release, version: str, model: str):
                """Download the ecoinvent release"""
                path = release.get_release(
                    version=version,
                    system_model=model,
                    release_type=ei.ReleaseType.ecospold,
                    extract=False,
                    fix_version=False
                )
                path = str(path)
                if not path.endswith(".7z") and os.path.exists(path + ".7z"):
                    path = path + ".7z"
                self.download_ready.emit(path)

        def __init__(self, parent=None):
            super().__init__(parent)
            self.ei_filepath = None

        def initializePage(self, context: dict):
            """Start the download thread"""
            self.thread.start(context["release"], context["version"], context["model"])
            self.thread.download_ready.connect(self.download_ready)

        def download_ready(self, filepath: str):
            """Handle the completion of the download"""
            self.ei_filepath = filepath

        def finalize(self, context: dict):
            """Store the downloaded file path in the context"""
            context["ei_filepath"] = self.ei_filepath

        def nextPage(self):
            """Proceed to the BiosphereSetupPage"""
            return EiWizard.BiosphereSetupPage

    class BiosphereSetupPage(widgets.ABWizardPage):
        """Wizard page to choose biosphere setup options"""
        title = "Biosphere setup"
        subtitle = "Choose whether to import the biosphere database or connect to an existing one"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.biosphere_choice = widgets.ABRadioButtonCollapser(self)
            self.biosphere_choice.buttonClicked.connect(lambda: self.completeChanged.emit())

            self.biosphere_choice.addOption(
                name="existing",
                label="Link to an existing biosphere",
                w=widgets.ABComboBox.get_database_combobox()
            )

            self.biosphere_choice.addOption(
                name="import",
                label="Import included biosphere",
                w=widgets.DatabaseNameEdit(database_preset="biosphere")
            )

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.biosphere_choice)
            self.setLayout(layout)

        def isComplete(self):
            """Check if a biosphere option has been selected"""
            return self.biosphere_choice.currentOption() is not None

        def initializePage(self, context: dict):
            """Start the biosphere installation thread"""
            if "version" in context:
                self.biosphere_choice.view("import").setText(f"biosphere-{context['version']}")

        def finalize(self, context: dict):
            """Store the selected biosphere option in the context"""
            if self.biosphere_choice.currentOption() == "existing":
                context["biosphere_name"] = self.biosphere_choice.view("existing").currentText()
            else:
                context["biosphere_name"] = self.biosphere_choice.view("import").text()

        def nextPage(self):
            """Proceed to the appropriate next page based on the biosphere choice"""
            if self.biosphere_choice.currentOption() == "existing":
                return EiWizard.EcoinventSetupPage
            else:
                return EiWizard.BiosphereInstallPage

    class BiosphereInstallPage(widgets.ABThreadedWizardPage):
        """Wizard page to install the biosphere database"""
        title = "Installing biosphere database"
        subtitle = "Installing bundled biosphere database into the project"

        class Thread(threading.ABThread):
            """Thread to handle the biosphere installation process"""
            def run_safely(self, ei_filepath: str, biosphere_name: str):
                """Install the biosphere database"""
                importer = Ecoinvent7zImporter(ei_filepath)
                importer.install_biosphere(biosphere_name)

        def initializePage(self, context: dict):
            """Start the biosphere installation thread"""
            self.thread.start(context["ei_filepath"], context["biosphere_name"])

        def nextPage(self):
            """Proceed to the MethodsSetupPage"""
            return EiWizard.MethodsSetupPage

    class MethodsSetupPage(widgets.ABWizardPage):
        """Wizard page to choose methods setup options"""
        title = "Methods setup"
        subtitle = "Choose whether to import methods from ecoinvent or from file"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.methods_choice = widgets.ABRadioButtonCollapser(self)
            self.methods_choice.buttonClicked.connect(lambda: self.completeChanged.emit())

            self.methods_choice.addOption(
                name="remote",
                label="Download methods from ecoinvent",
                w=QtWidgets.QWidget()
            )

            self.methods_choice.addOption(
                name="local",
                label="Import methods from file",
                w=widgets.ABFileSelector(filter="*.xlsx")
            )

            self.methods_choice.addOption(
                name="skip",
                label="Don't import methods",
                w=QtWidgets.QWidget()
            )

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.methods_choice)

            self.setLayout(layout)

        def finalize(self, context: dict):
            """Store the selected methods option in the context"""
            if self.methods_choice.currentOption() == "remote":
                file = ei.get_excel_lcia_file_for_version(context["release"], context["version"])
                context["methods_filepath"] = str(file)
            if self.methods_choice.currentOption() == "local":
                context["methods_filepath"] = self.methods_choice.view("local").text()

        def isComplete(self):
            """Check if a methods option has been selected"""
            return self.methods_choice.currentOption() is not None

        def nextPage(self):
            """Proceed to the appropriate next page based on the methods choice"""
            if self.methods_choice.currentOption() == "remote" or self.methods_choice.currentOption() == "local":
                return EiWizard.MethodsInstallPage
            else:
                return EiWizard.EcoinventSetupPage

    class MethodsInstallPage(widgets.ABThreadedWizardPage):
        """Wizard page to install the selected methods"""
        title = "Installing methods"
        subtitle = "Installing selected methods and linking to the biosphere"

        class Thread(threading.ABThread):
            """Thread to handle the methods installation process"""
            def run_safely(self, methods_filepath: str, biosphere_name: str):
                """Install the methods and link to the biosphere"""
                importer = EcoinventLCIAImporter.setup_with_ei_excel(methods_filepath)
                importer.set_biosphere(biosphere_name)
                importer.apply_strategies()

                signals.method.blockSignals(True)
                signals.meta.blockSignals(True)

                old = bd.methods.deserialize()
                importer.write_methods(overwrite=True)

                signals.method.blockSignals(False)
                signals.meta.blockSignals(False)

                signals.meta.methods_changed.emit(deepcopy(bd.methods.data), old)

        def initializePage(self, context: dict):
            """Start the methods installation thread"""
            self.thread.start(context["methods_filepath"], context["biosphere_name"])

        def nextPage(self):
            """Proceed to the EcoinventSetupPage"""
            return EiWizard.EcoinventSetupPage

    class EcoinventSetupPage(widgets.ABWizardPage):
        """Wizard page to set up the ecoinvent database"""
        title = "Ecoinvent setup"
        subtitle = "Choose name for ecoinvent database"

        def __init__(self, parent=None):
            super().__init__(parent)
            self.database_name = widgets.DatabaseNameEdit(database_preset="ecoinvent")
            self.database_name.textChanged.connect(self.completeChanged)

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.database_name)
            self.setLayout(layout)

        def isComplete(self):
            """Check if a database name has been entered"""
            return bool(self.database_name.text())

        def initializePage(self, context: dict):
            """Start the biosphere installation thread"""
            if "version" in context and "model" in context:
                self.database_name.setText(f"ecoinvent-{context['version']}-{context['model']}")

        def finalize(self, context: dict):
            """Store the database name in the context"""
            context["database_name"] = self.database_name.text()

        def nextPage(self):
            """Proceed to the EcoinventInstallPage"""
            return EiWizard.EcoinventInstallPage

    class EcoinventInstallPage(widgets.ABThreadedWizardPage):
        """Wizard page to install the ecoinvent database"""
        title = "Installing ecoinvent"
        subtitle = "Installing ecoinvent database into the project"

        class Thread(threading.ABThread):
            """Thread to handle the ecoinvent installation process"""
            def run_safely(self, ei_filepath: str, database_name: str, biosphere_name: str):
                """Install the ecoinvent database"""
                importer = Ecoinvent7zImporter(ei_filepath)
                importer.install_ecoinvent(database_name, biosphere_name)

                # Run migrations after installation
                if len(bi.migrations) == 0:
                    ab_create_core_migrations()

        def initializePage(self, context: dict):
            """Start the ecoinvent installation thread"""
            self.thread.start(
                context["ei_filepath"],
                context["database_name"],
                context["biosphere_name"]
            )

    pages = [
        RemoteOrLocalPage, LocalSelectPage, LoginPage, EcoinventVersionPage, EcoinventDownloadPage, BiosphereSetupPage,
        BiosphereInstallPage, MethodsSetupPage, MethodsInstallPage, EcoinventSetupPage, EcoinventInstallPage
    ]



from typing import Any

from activity_browser import signals
from activity_browser.actions.base import ABAction, exception_dialogs
from bw2data import get_activity
from bw2data.parameters import (ActivityParameter, Group,
                                                     GroupDependency,
                                                     parameters)
from activity_browser.ui.icons import qicons


class ParameterDelete(ABAction):
    """
    ABAction to delete an existing parameter.
    """

    icon = qicons.delete
    text = "Delete parameter..."

    @staticmethod
    @exception_dialogs
    def run(parameter: Any):
        if isinstance(parameter, ActivityParameter):
            db = parameter.database
            code = parameter.code
            amount = (
                ActivityParameter.select()
                .where(
                    (ActivityParameter.database == db)
                    & (ActivityParameter.code == code)
                )
                .count()
            )

            if amount > 1:
                parameter.delete_instance()
            else:
                group = parameter.group
                act = get_activity((db, code))
                parameters.remove_from_group(group, act)
                # Also clear the group if there are no more parameters in it

                if (
                    not ActivityParameter.select()
                    .where(ActivityParameter.group == group)
                    .exists()
                ):
                    Group.delete().where(Group.name == group).execute()
                    GroupDependency.delete().where(
                        GroupDependency.group == group
                    ).execute()
        else:
            parameter.delete_instance()
        # After deleting things, recalculate and signal changes
        parameters.recalculate()

        # No fire when everything is still fresh after recalculation, so need to fire manually to be sure everything is
        # updated correctly.
        signals.parameter.recalculated.emit()



from tqdm.std import *
from activity_browser.mod.patching import patch_attribute, patched
from qtpy.QtCore import QObject, SignalInstance, Signal


@patch_attribute(tqdm, "update")
def update(self, n=1):
    patched[tqdm]["update"](self, n)
    try:
        qt_tqdm.updated.emit(int(self.n/self.total * 100), self.desc)
    except (TypeError, ZeroDivisionError):
        # Handle case where total is None or zero
        pass


@patch_attribute(tqdm, "close")
def close(self):
    patched[tqdm]["close"](self)
    qt_tqdm.updated.emit(100, self.desc)


class QtTqdm(QObject):
    updated: SignalInstance = Signal(int, str)


qt_tqdm = QtTqdm()



from typing import Tuple

from qtpy import QtCore, QtGui, QtWidgets

from activity_browser import actions, application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import commontasks as bc
from activity_browser.mod import bw2data as bd
from bw2data.parameters import ActivityParameter
from activity_browser.ui.icons import qicons

PARAMETER_STRINGS = (
    "Project: Available to all other parameters",
    "Database: Available to Database and Activity parameters of the same database",
    "Activity: Available to Activity and exchange parameters within the group",
)
PARAMETER_FIELDS = (
    ("name", "amount"),
    ("name", "amount", "database"),
    ("name", "amount"),
)


class ParameterNew(ABAction):
    """
    ABAction to create a new Parameter. Opens the ParameterWizard, returns if the wizard is canceled. Else,
    checks whether the name is valid, and then instructs the ParameterController to put the new parameter in the
    right group.
    """

    icon = qicons.add
    text = "New parameter..."

    @staticmethod
    @exception_dialogs
    def run(activity_key: Tuple[str, str]):
        # instantiate the ParameterWizard
        wizard = ParameterWizard(activity_key, application.main_window)

        # return if the wizard is canceled
        if wizard.exec_() != QtWidgets.QWizard.Accepted:
            return

        # gather wizard variables
        selection = wizard.selected
        data = wizard.param_data

        # check whether the name is valid, otherwise return
        name = data.get("name")
        if name[0] in ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "-", "#"):
            error = QtWidgets.QErrorMessage()
            error.showMessage(
                "<p>Parameter names must not start with a digit, hyphen, or hash character</p>"
            )
            error.exec_()
            return

        # select the right group and instruct the controller to create the parameter there
        if selection == 0:
            bd.parameters.new_project_parameters([data])
        elif selection == 1:
            db = data.pop("database")
            bd.parameters.new_database_parameters([data], db)
        elif selection == 2:
            group = data.pop("group")
            bd.parameters.new_activity_parameters([data], group)


class ParameterWizard(QtWidgets.QWizard):
    complete = QtCore.Signal(str, str, str)

    def __init__(self, key: tuple, parent=None):
        super().__init__(parent)

        self.key = key
        self.pages = (
            SelectParameterTypePage(self),
            CompleteParameterPage(self),
        )
        for i, p in enumerate(self.pages):
            self.setPage(i, p)

    @property
    def selected(self) -> int:
        return self.pages[0].selected

    @property
    def param_data(self) -> dict:
        data = {field: self.field(field) for field in PARAMETER_FIELDS[self.selected]}
        if self.selected == 2:
            data["group"] = self._get_group()
            data["database"] = self.key[0]
            data["code"] = self.key[1]
        return data

    def _get_group(self):
        query = (ActivityParameter.database == self.key[0]) & (
            ActivityParameter.code == self.key[1]
        )

        if not ActivityParameter.select().where(query).count():
            actions.ParameterNewAutomatic.run([self.key])

        return ActivityParameter.get(query).group


class SelectParameterTypePage(QtWidgets.QWizardPage):
    def __init__(self, parent):
        super().__init__(parent)
        self.setTitle("Select the type of parameter to create.")

        self.key = parent.key

        layout = QtWidgets.QVBoxLayout()
        box = QtWidgets.QGroupBox("Types:")
        # Explicitly set the stylesheet to avoid parent classes overriding
        box.setStyleSheet(
            "QGroupBox {border: 1px solid gray; border-radius: 5px; margin-top: 7px; margin-bottom: 7px; padding: 0px}"
            "QGroupBox::title {top:-7 ex;left: 10px; subcontrol-origin: border}"
        )
        box_layout = QtWidgets.QVBoxLayout()
        self.button_group = QtWidgets.QButtonGroup()
        self.button_group.setExclusive(True)
        for i, s in enumerate(PARAMETER_STRINGS):
            button = QtWidgets.QRadioButton(s)
            self.button_group.addButton(button, i)
            box_layout.addWidget(button)
        # If we have a complete key, pre-select the activity parameter btn.
        if all(self.key):
            self.button_group.button(2).setChecked(True)
        elif self.key[0] != "":
            # default to database parameter is we have something.
            self.button_group.button(2).setEnabled(False)
            self.button_group.button(1).setChecked(True)
        else:
            # If we don't have a complete key, we can't create an activity parameter
            self.button_group.button(2).setEnabled(False)
            self.button_group.button(0).setChecked(True)
        box.setLayout(box_layout)
        layout.addWidget(box)
        self.setLayout(layout)

    @property
    def selected(self) -> int:
        return self.button_group.checkedId()


class CompleteParameterPage(QtWidgets.QWizardPage):
    def __init__(self, parent):
        super().__init__(parent)
        self.setTitle("Fill out required values for the parameter")
        self.parent = parent

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        box = QtWidgets.QGroupBox("Data:")
        box.setStyleSheet(
            "QGroupBox {border: 1px solid gray; border-radius: 5px; margin-top: 7px; margin-bottom: 7px; padding: 0px}"
            "QGroupBox::title {top:-7 ex;left: 10px; subcontrol-origin: border}"
        )
        grid = QtWidgets.QGridLayout()
        box.setLayout(grid)
        layout.addWidget(box)

        self.key = parent.key

        self.name_label = QtWidgets.QLabel("Name:")
        self.name = QtWidgets.QLineEdit()
        grid.addWidget(self.name_label, 0, 0)
        grid.addWidget(self.name, 0, 1)
        self.amount_label = QtWidgets.QLabel("Amount:")
        self.amount = QtWidgets.QLineEdit()
        locale = QtCore.QLocale(QtCore.QLocale.English)
        locale.setNumberOptions(QtCore.QLocale.RejectGroupSeparator)
        validator = QtGui.QDoubleValidator()
        validator.setLocale(locale)
        self.amount.setValidator(validator)
        grid.addWidget(self.amount_label, 1, 0)
        grid.addWidget(self.amount, 1, 1)
        self.database_label = QtWidgets.QLabel("Database:")
        self.database = QtWidgets.QComboBox()
        grid.addWidget(self.database_label, 2, 0)
        grid.addWidget(self.database, 2, 1)

        # Register fields for all possible values
        self.registerField("name*", self.name)
        self.registerField("amount", self.amount)
        self.registerField("database", self.database, "currentText")

    def initializePage(self) -> None:
        self.amount.setText("1.0")
        if self.parent.selected == 0:
            self.name.clear()
            self.database.setHidden(True)
            self.database_label.setHidden(True)
        elif self.parent.selected == 1:
            self.name.clear()
            self.database.clear()
            dbs = list(bd.databases)
            self.database.insertItems(0, dbs)
            if self.key[0] in dbs:
                self.database.setCurrentIndex(dbs.index(self.key[0]))
            self.database.setHidden(False)
            self.database_label.setHidden(False)
        elif self.parent.selected == 2:
            self.name.clear()
            self.database.setHidden(True)
            self.database_label.setHidden(True)



from logging import getLogger

from activity_browser import bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd

log = getLogger(__name__)


class CSChangeFunctionalUnit(ABAction):
    """
    Updates the functional unit amount for a specific inventory item in a calculation setup.

    This method modifies the amount of a functional unit in the inventory of a given calculation setup
    and saves the updated setup.

    Args:
        cs_name (str): The name of the calculation setup to modify.
        index (int): The index of the inventory item within the calculation setup.
        amount (float): The new amount to set for the functional unit.

    Steps:
    - Retrieve the calculation setup by its name.
    - Extract the key of the inventory item at the specified index.
    - Update the amount for the specified inventory item.
    - Serialize and save the updated calculation setup.

    Raises:
        Exception: If an error occurs during the process, it is handled by the `exception_dialogs` decorator.
    """
    text = "Add Functional Unit to Calculation Setup"

    @staticmethod
    @exception_dialogs
    def run(cs_name: str, index: int, amount: float):
        calculation_setup = bd.calculation_setups[cs_name]

        key = list(calculation_setup['inv'][index].keys())[0]
        calculation_setup['inv'][index][key] = amount

        bd.calculation_setups.serialize()



from typing import List

from qtpy import QtWidgets

import bw2data as bd
import bw_functional as bf

from bw2data.parameters import (ActivityParameter, Group,
                                                     GroupDependency,
                                                     parameters)

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ActivityDelete(ABAction):
    """
    ABAction to delete one or multiple activities if supplied by activity keys. Will check if an activity has any
    downstream processes and ask the user whether they want to continue if so. Exchanges from any downstream processes
    will be removed
    """

    icon = qicons.delete
    text = "Delete ***"

    @staticmethod
    @exception_dialogs
    def run(activity_keys: List[tuple]):
        # retrieve activity objects from the controller using the provided keys
        activities = [bd.get_activity(key) for key in activity_keys]

        warnings = [f"Are you certain you want to delete {len(activities)} activity/activities?", ""]

        if any(len(act.upstream()) > 0 for act in activities):
            warnings.append("One or more of the activities you are trying to delete have consumers")

        if any([act for act in activities if isinstance(act, bf.Process)]):
            warnings.append("Products of processes will be removed as well")

        warning_text = "\n".join(warnings)

        # alert the user
        choice = QtWidgets.QMessageBox.warning(
            application.main_window,
            "Deleting activity/activities",
            warning_text,
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No,
        )

        # return if the user cancels
        if choice == QtWidgets.QMessageBox.No:
            return

        for act in activities:
            db, code = act.key

            try:
                group_name = ActivityParameter.get(
                    (ActivityParameter.database == db)
                    & (ActivityParameter.code == code)
                ).group

                # remove activity parameters from its group
                parameters.remove_from_group(group_name, act)

                # Also clear the group if there are no more parameters in it
                if (
                    not ActivityParameter.select()
                    .where(ActivityParameter.group == group_name)
                    .exists()
                ):
                    Group.get(Group.name == group_name).delete_instance()
                    GroupDependency.delete().where(
                        GroupDependency.group == group_name
                    ).execute()
            except ActivityParameter.DoesNotExist:
                # no parameters found for this activity
                pass

            # Included in bw2data as of 4.1
            # act.upstream().delete()

            act.delete()



from typing import Any

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class ParameterUncertaintyModify(ABAction):
    """
    ABAction to modify the uncertainty of an existing parameter.
    """

    icon = qicons.edit
    text = "Modify parameter uncertainty"

    @staticmethod
    @exception_dialogs
    def run(parameter: Any, uncertainty_dict: dict):
        parameter.data.update(uncertainty_dict)
        parameter.save()
        bd.parameters.recalculate()



from qtpy import QtWidgets, QtGui

def horizontal_line():
    line = QtWidgets.QFrame()
    line.setFrameShape(QtWidgets.QFrame.HLine)
    line.setFrameShadow(QtWidgets.QFrame.Sunken)
    return line


def vertical_line():
    line = QtWidgets.QFrame()
    line.setFrameShape(QtWidgets.QFrame.VLine)
    line.setFrameShadow(QtWidgets.QFrame.Sunken)
    return line


def header(text):
    label = QtWidgets.QLabel(text)

    bold_font = QtGui.QFont()
    bold_font.setBold(True)
    bold_font.setPointSize(12)

    label.setFont(bold_font)
    return label


from logging import getLogger

import bw2data as bd

from activity_browser.actions.base import ABAction, exception_dialogs

log = getLogger(__name__)


class CSAddImpactCategory(ABAction):
    text = "Add Impact Category to Calculation Setup"

    @staticmethod
    @exception_dialogs
    def run(cs_name: str, ic_names: list[str]):
        calculation_setup = bd.calculation_setups[cs_name]

        calculation_setup['ia'] += ic_names

        bd.calculation_setups[cs_name] = calculation_setup
        bd.calculation_setups.serialize()



# -*- coding: utf-8 -*-
from ast import literal_eval
from pathlib import Path
from typing import List, Union
from logging import getLogger

import openpyxl
import pandas as pd

from .utils import SUPERSTRUCTURE

log = getLogger(__name__)


def convert_tuple_str(x):
    try:
        return literal_eval(x)
    except (ValueError, SyntaxError) as e:
        return x


def get_sheet_names(document_path: Union[str, Path]) -> List[str]:
    try:
        wb = openpyxl.load_workbook(filename=document_path, read_only=True)
        return wb.sheetnames
    except UnicodeDecodeError as e:
        log.error("Given document uses an unknown encoding: {}".format(e))


def get_header_index(document_path: Union[str, Path], import_sheet: int):
    """Retrieves the line index for the column headers, will raise an
    exception if not found in the first 10 rows.
    """
    try:
        wb = openpyxl.load_workbook(filename=document_path, read_only=True)
        sheet = wb.worksheets[import_sheet]
        for i in range(10):
            value = sheet.cell(i + 1, 1).value
            if isinstance(value, str):
                wb.close()
                return i
    except IndexError as e:
        wb.close()
        raise IndexError("Expected headers not found in file").with_traceback(
            e.__traceback__
        )
    except UnicodeDecodeError as e:
        log.error("Given document uses an unknown encoding: {}".format(e))
        wb.close()
    raise ValueError("Could not find required headers in given document sheet.")


def valid_cols(name: str) -> bool:
    """Callable which evaluates if a specific column should be used."""
    return False if str(name).startswith("#") else True


def import_from_excel(
    document_path: Union[str, Path], import_sheet: int = 1
) -> pd.DataFrame:
    """Import all of the exchanges and their scenario amounts from a given
    document and sheet index.

    The default index chosen represents the second sheet (first after the
    'information' sheet).

    Any '*' character used at the start of a row or will cause that row
    to be excluded from the import.
    A '#' character at the start of a column will cause that column to be
    excluded from the import.

    'usecols' is used to exclude specific columns from the excel document.
    'comment' is used to exclude specific rows from the excel document.
    """
    data = pd.DataFrame({})
    try:
        header_idx = get_header_index(document_path, import_sheet)
        data = pd.read_excel(
            document_path,
            sheet_name=import_sheet,
            header=header_idx,
            usecols=valid_cols,
            comment="*",
            na_values="",
            keep_default_na=False,
            engine="openpyxl",
        )
        diff = SUPERSTRUCTURE.difference(data.columns)
        if not diff.empty:
            raise ValueError(
                "Missing required column(s) for superstructure: {}".format(
                    diff.to_list()
                )
            )

        # Convert specific columns that may have tuples as strings
        columns = ["from categories", "from key", "to categories", "to key"]
        data.loc[:, columns] = data[columns].map(convert_tuple_str)
    except:
        # skip the error checks here, these now occur in the calling layout.tabs.LCA_setup module
        pass
    return data



from logging import getLogger
from qtpy import QtCore, QtGui, QtWidgets

from activity_browser import application

log = getLogger(__name__)


class ABAction:
    icon = QtGui.QIcon()
    text: str = None
    tooltip: str = None

    @staticmethod
    def run(*args, **kwargs):
        raise NotImplementedError

    @classmethod
    def triggered(cls, *args, **kwargs):
        args = [arg if not callable(arg) else arg() for arg in args]
        kwargs = {k: v if not callable(v) else v() for k, v in kwargs.items()}

        cls.run(*args, **kwargs)

    @classmethod
    def get_QAction(cls, *args, parent=None, text=None, enabled=True, **kwargs) -> QtWidgets.QAction:
        text = text or cls.text
        action = QtWidgets.QAction(cls.icon, text, parent, enabled=enabled)
        action.setToolTip(cls.tooltip)

        action.triggered.connect(lambda: cls.triggered(*args, **kwargs))

        return action

    @classmethod
    def get_QButton(cls, *args, **kwargs):
        """Convenience function to return a button that has this ABAction as default action."""
        button = QtWidgets.QPushButton(
            cls.icon,
            cls.text
        )
        button.clicked.connect(lambda x: cls.triggered(*args, **kwargs))
        return button


def exception_dialogs(func):
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            if not hasattr(e, "dialog_flag"):
                setattr(e, "dialog_flag", True)
                QtWidgets.QMessageBox.critical(
                    application.main_window,
                    f"An error occurred: {type(e).__name__}",
                    f"An error occurred, check the logs for more information \n\n {str(e)}",
                    QtWidgets.QMessageBox.Ok,
                )
            raise e

    return wrapper



import itertools
from typing import Iterable, Tuple
from logging import getLogger

import pandas as pd
import numpy as np
from asteval import Interpreter
from peewee import DoesNotExist

from qtpy import QtWidgets
from qtpy.QtCore import QModelIndex, Slot

from bw2data.parameters import ActivityParameter, DatabaseParameter, Group, ProjectParameter

from activity_browser import actions, signals, application, bwutils
from activity_browser.mod import bw2data as bd
from activity_browser.ui.wizards import UncertaintyWizard

from .base import BaseTreeModel, EditablePandasModel, TreeItem, PandasModel

log = getLogger(__name__)


class BaseParameterModel(EditablePandasModel):
    COLUMNS = []
    UNCERTAINTY = ["uncertainty type", "loc", "scale", "shape", "minimum", "maximum"]

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.param_col = 0
        self.comment_col = 0
        self.dataChanged.connect(self.edit_single_parameter)
        self.set_read_only(False)

        signals.project.changed.connect(self.sync)
        signals.parameter.changed.connect(self.sync)
        signals.parameter.recalculated.connect(self.sync)

    def get_parameter(self, proxy: QModelIndex) -> object:
        idx = self.proxy_to_source(proxy)
        return self._dataframe.iat[idx.row(), self.param_col]

    def get_key(self, *args) -> tuple:
        """Use this to build a (partial) key for the current index."""
        return "", ""

    def get_group(self, *args) -> str:
        """Retrieve the group of the parameter currently selected."""
        return "project"

    @classmethod
    def parse_parameter(cls, parameter) -> dict:
        """Take the given Parameter object and extract data for a single
        row in the table dataframe

        If the parameter has uncertainty data, include this as well.
        """
        row = {key: getattr(parameter, key, "") for key in cls.COLUMNS}
        data = getattr(parameter, "data", {})
        row.update(cls.extract_uncertainty_data(data))
        row["parameter"] = parameter
        row["comment"] = data.get("comment", "")
        return row

    @classmethod
    def columns(cls) -> list:
        """Combine COLUMNS, UNCERTAINTY and add 'parameter'."""
        return cls.COLUMNS + cls.UNCERTAINTY + ["parameter"]

    @classmethod
    def extract_uncertainty_data(cls, data: dict) -> dict:
        """This helper function can be used to extract specific uncertainty
        columns from the parameter data

        See:
        https://2.docs.brightway.dev/intro.html#storing-uncertain-values
        https://stats-arrays.readthedocs.io/en/latest/#mapping-parameter-array-columns-to-uncertainty-distributions
        """
        row = {key: data.get(key) for key in cls.UNCERTAINTY}
        return row

    @Slot(QModelIndex, name="editSingleParameter")
    def edit_single_parameter(self, index: QModelIndex) -> None:
        """Take the index and update the underlying brightway Parameter."""
        param = self.get_parameter(index)
        field = self._dataframe.columns[index.column()]

        try:
            actions.ParameterModify.run(param, field, index.data())
        except Exception as e:
            QtWidgets.QMessageBox.warning(
                application.main_window,
                "Could not save changes",
                str(e),
                QtWidgets.QMessageBox.Ok,
                QtWidgets.QMessageBox.Ok,
            )

    @Slot(QModelIndex, name="startRenameParameter")
    def handle_parameter_rename(self, proxy: QModelIndex) -> None:
        group = self.get_group(proxy)
        param = self.get_parameter(proxy)

        actions.ParameterRename.run(param)

    def delete_parameter(self, proxy: QModelIndex) -> None:
        param = self.get_parameter(proxy)
        actions.ParameterDelete.run(param)

    @Slot(name="modifyParameterUncertainty")
    def modify_uncertainty(self, proxy: QModelIndex) -> None:
        param = self.get_parameter(proxy)
        wizard = UncertaintyWizard(param, self.parent())
        wizard.show()

    @Slot(name="unsetParameterUncertainty")
    def remove_uncertainty(self, proxy: QModelIndex) -> None:
        param = self.get_parameter(proxy)
        actions.ParameterUncertaintyRemove.run(param)

    def handle_double_click(self, proxy: QModelIndex) -> None:
        column = proxy.column()
        if self._dataframe.columns[column] in BaseParameterModel.UNCERTAINTY:
            self.modify_uncertainty(proxy)
        elif self._dataframe.columns[column] == "name":
            self.handle_parameter_rename(proxy)


class ProjectParameterModel(BaseParameterModel):
    COLUMNS = ["name", "amount", "formula", "comment"]

    def sync(self) -> None:
        data = [self.parse_parameter(p) for p in ProjectParameter.select()]
        self._dataframe = pd.DataFrame(data, columns=self.columns())
        self.param_col = self._dataframe.columns.get_loc("parameter")
        self.comment_col = self._dataframe.columns.get_loc("comment")
        self.updated.emit()

    @staticmethod
    def get_usable_parameters() -> Iterable[list]:
        return ([k, v, "project"] for k, v in ProjectParameter.static().items())

    @staticmethod
    def get_interpreter() -> Interpreter:
        interpreter = Interpreter()
        interpreter.symtable.update(ProjectParameter.static())
        return interpreter


class DatabaseParameterModel(BaseParameterModel):
    COLUMNS = ["name", "amount", "formula", "database", "comment"]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.db_col = 0

    def sync(self) -> None:
        data = [self.parse_parameter(p) for p in DatabaseParameter.select()]
        self._dataframe = pd.DataFrame(data, columns=self.columns())
        self.db_col = self._dataframe.columns.get_loc("database")
        self.param_col = self._dataframe.columns.get_loc("parameter")
        self.comment_col = self._dataframe.columns.get_loc("comment")
        self.updated.emit()

    def get_key(self, proxy: QModelIndex = None) -> tuple:
        return self.get_database(proxy), ""

    def get_group(self, proxy: QModelIndex = None) -> str:
        """Retrieve the group of the activity currently selected."""
        return self.get_database(proxy)

    @staticmethod
    def get_usable_parameters():
        """Include the project parameters, and generate database parameters."""
        project = ProjectParameterModel.get_usable_parameters()
        database = (
            [p.name, p.amount, "database ({})".format(p.database)]
            for p in DatabaseParameter.select()
        )
        return itertools.chain(project, database)

    def get_database(self, proxy: QModelIndex = None) -> str:
        """Return the database name of the parameter currently selected."""
        idx = self.proxy_to_source(proxy or self.parent().currentIndex())
        return self._dataframe.iat[idx.row(), self.db_col]

    def get_interpreter(self) -> Interpreter:
        """Take the interpreter from the ProjectParameterTable and add
        (potentially overwriting) all database symbols for the selected index.
        """
        interpreter = ProjectParameterModel.get_interpreter()
        db_name = self.get_database()
        interpreter.symtable.update(DatabaseParameter.static(db_name))
        return interpreter


class ActivityParameterModel(BaseParameterModel):
    COLUMNS = [
        "name",
        "amount",
        "formula",
        "product",
        "activity",
        "location",
        "group",
        "order",
        "key",
        "comment",
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.group_col = 0
        self.key_col = 0
        self.order_col = 0

    def sync(self) -> None:
        """Build a dataframe using the ActivityParameters set in brightway"""
        generate = (
            self.parse_parameter(p)
            for p in (
                ActivityParameter.select(ActivityParameter, Group.order)
                .join(Group, on=(ActivityParameter.group == Group.name))
                .namedtuples()
            )
        )
        data = [x for x in generate if "key" in x]
        self._dataframe = pd.DataFrame(data, columns=self.columns())
        # Convert the 'order' column from list into string
        self._dataframe["order"] = self._dataframe["order"].apply(", ".join)
        self.group_col = self._dataframe.columns.get_loc("group")
        self.param_col = self._dataframe.columns.get_loc("parameter")
        self.key_col = self._dataframe.columns.get_loc("key")
        self.order_col = self._dataframe.columns.get_loc("order")
        self.comment_col = self._dataframe.columns.get_loc("comment")
        self.updated.emit()

    @classmethod
    def parse_parameter(cls, parameter) -> dict:
        """Override the base method to add more steps."""
        row = super().parse_parameter(parameter)
        # Combine the 'database' and 'code' fields of the parameter into a 'key'
        row["key"] = (parameter.database, parameter.code)
        try:
            act = bd.get_activity(row["key"])
        except:
            # Can occur if an activity parameter exists for a removed activity.
            log.info(
                "Activity {} no longer exists, removing parameter.".format(row["key"])
            )
            actions.ParameterClearBroken.run(parameter)
            return {}
        row["product"] = act.get("reference product") or act.get("name")
        row["activity"] = act.get("name")
        row["location"] = act.get("location", "unknown")
        # Replace the namedtuple with the actual ActivityParameter
        row["parameter"] = ActivityParameter.get_by_id(parameter.id)
        return row

    def get_activity_groups(self, proxy, ignore_groups: list = None) -> Iterable[str]:
        """Helper method to look into the Group and determine which if any
        other groups the current activity can depend on
        """
        db = self.get_key(proxy)[0]
        ignore_groups = ignore_groups or []
        return (
            param.group
            for param in (
                ActivityParameter.select(ActivityParameter.group)
                .where(ActivityParameter.database == db)
                .distinct()
            )
            if param.group not in ignore_groups
        )

    @staticmethod
    def get_usable_parameters():
        """Include all types of parameters.

        NOTE: This method does not take into account which formula is being
        edited, and therefore does not restrict which database or activity
        parameters are returned.
        """
        database = DatabaseParameterModel.get_usable_parameters()
        activity = (
            [p.name, p.amount, "activity ({})".format(p.group)]
            for p in ActivityParameter.select()
        )
        return itertools.chain(database, activity)

    def get_group(self, proxy: QModelIndex = None) -> str:
        """Retrieve the group of the activity currently selected."""
        proxy = proxy or self.parent().currentIndex()
        idx = self.proxy_to_source(proxy)
        return self._dataframe.iat[idx.row(), self.group_col]

    def get_interpreter(self) -> Interpreter:
        interpreter = Interpreter()
        group = self.get_group(self.parent().currentIndex())
        interpreter.symtable.update(ActivityParameter.static(group, full=True))
        return interpreter

    def get_key(self, proxy: QModelIndex) -> tuple:
        index = self.proxy_to_source(proxy)
        return self._dataframe.iat[index.row(), self.key_col]


class ParameterItem(TreeItem):
    @classmethod
    def build_header(cls, header: str, parent: TreeItem) -> "ParameterItem":
        item = cls([header, "", "", ""], parent)
        parent.appendChild(item)
        return item

    @classmethod
    def build_item(cls, param, parent: TreeItem) -> "ParameterItem":
        """Depending on the parameter type, the group is changed, defaults to
        'project'.

        For Activity parameters, use a 'header' item as parent, create one
        if it does not exist.
        """
        group = "project"
        if hasattr(param, "code") and hasattr(param, "database"):
            database = "database - {}".format(str(param.database))
            if database not in [x.data(0) for x in parent.children]:
                cls.build_header(database, parent)
            parent = next(x for x in parent.children if x.data(0) == database)
            group = getattr(param, "group")
        elif hasattr(param, "database"):
            group = param.database

        item = cls(
            [
                getattr(param, "name", ""),
                group,
                getattr(
                    param, "amount", 1.0
                ),  # set to 1 instead of 0 as division by 0 causes problems
                getattr(param, "formula", ""),
            ],
            parent,
        )

        # If the variable is found, we're working on an activity parameter
        if "database" in locals():
            cls.build_exchanges(param, item)

        parent.appendChild(item)
        return item

    @classmethod
    def build_exchanges(cls, act_param, parent: TreeItem) -> None:
        """Take the given activity parameter, retrieve the matching activity
        and construct tree-items for each exchange with a `formula` field.
        """
        act = bd.get_activity((act_param.database, act_param.code))

        for exc in [exc for exc in act.exchanges() if "formula" in exc]:
            try:
                act_input = bd.get_activity(exc.input)
                item = cls(
                    [
                        act_input.get("name"),
                        parent.data(1),
                        exc.amount,
                        exc.get("formula"),
                    ],
                    parent,
                )
                parent.appendChild(item)
            except DoesNotExist as e:
                # The exchange is coming from a deleted database, remove it
                log.warning(f"Broken exchange: {exc}, removing.")
                actions.ExchangeDelete.run([exc])


class ParameterTreeModel(BaseTreeModel):
    """
    Ordering and foldouts as follows:
    - Project parameters:
        - All 'root' objects
        - No children
    - Database parameters:
        - All 'root' objects
        - No children
    - Activity parameters:
        - Never root objects.
        - Placed under simple 'database' root objects
        - Exchanges as children
    - Exchange parameters:
        - Never root objects
        - Children of relevant activity parameter
        - No children
    """

    HEADERS = ["Name", "Group", "Amount", "Formula"]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.root = ParameterItem.build_root(self.HEADERS)
        self.setup_model_data()

    def setup_model_data(self) -> None:
        """First construct the root, then process the data."""
        for param in self._data.get("project", []):
            ParameterItem.build_item(param, self.root)
        for param in self._data.get("database", []):
            ParameterItem.build_item(param, self.root)
        for param in self._data.get("activity", []):
            try:
                _ = bd.get_activity((param.database, param.code))
            except:
                continue
            ParameterItem.build_item(param, self.root)

    def sync(self, *args, **kwargs) -> None:
        self.beginResetModel()
        self.root.clear()
        self.endResetModel()
        self._data.update(
            {
                "project": ProjectParameter.select().iterator(),
                "database": DatabaseParameter.select().iterator(),
                "activity": ActivityParameter.select().iterator(),
            }
        )
        self.setup_model_data()
        self.updated.emit()


class ScenarioModel(PandasModel):
    HEADERS = ["Name", "Group", "default"]
    MATCH_COLS = ["Name", "Group"]

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        signals.project.changed.connect(self.sync)
        signals.parameter.changed.connect(self.rebuild_table)

    @Slot(name="doCleanSync")
    def sync(self, df: pd.DataFrame = None, include_default: bool = True) -> None:
        """Construct the dataframe from the existing parameters, if ``df``
        is given, perform a merge to possibly include additional columns.
        """
        data = [p[:3] for p in bwutils.utils.Parameters.from_bw_parameters()]
        if not isinstance(df, pd.DataFrame):
            self._dataframe = pd.DataFrame(data, columns=self.HEADERS).set_index("Name")
        else:
            required = set(self.MATCH_COLS)
            if not required.issubset(df.columns):
                raise ValueError(
                    "The given dataframe does not contain required columns: {}".format(
                        required.difference(df.columns)
                    )
                )
            assert df.columns.get_loc("Group") == 1
            if isinstance(include_default, bool) and include_default:
                new_df = pd.DataFrame(data, columns=self.HEADERS)
                if "default" in df.columns:
                    df.drop(columns="default", inplace=True)
                self._dataframe = self._perform_merge(new_df, df).set_index("Name")
            else:
                # Now we're gonna need to ensure that the dataframe is of
                # the same size
                assert (
                    len(data) >= df.shape[0]
                ), "Too many parameters found, not possible."
                missing = len(data) - df.shape[0]
                if missing != 0:
                    nan_data = pd.DataFrame(
                        index=pd.RangeIndex(missing), columns=df.columns
                    )
                    df = pd.concat([df, nan_data], ignore_index=True)
                self._dataframe = df.set_index("Name")
        self.updated.emit()

    @classmethod
    def _perform_merge(cls, left: pd.DataFrame, right: pd.DataFrame) -> pd.DataFrame:
        """There are three kinds of actions that can occur: adding new columns,
        updating values in matching columns, and a combination of the two.

        ``left`` dataframe always determines the row-size of the resulting
        dataframe.
        Any `NaN` values in the new columns in ``right`` will be replaced
        with values from the `default` column from ``left``.
        """
        right_columns = right.drop(columns=cls.MATCH_COLS).columns
        matching = right_columns.intersection(left.columns)
        if not matching.empty:
            # Replace values and drop the matching columns
            left[matching] = right[matching]
            right.drop(columns=matching, inplace=True)
            if right.drop(columns=cls.MATCH_COLS).columns.any():
                # Merge the remaining columns
                df = left.merge(right, how="left", on=cls.MATCH_COLS)
            else:
                df = left
        else:
            df = left.merge(right, how="left", on=cls.MATCH_COLS)
        # Now go over the non-standard columns and see if there are any
        # missing values.
        new_cols = df.drop(columns=cls.HEADERS).columns
        missing = new_cols[df[new_cols].isna().any()]
        if not missing.empty:
            idx = missing.append(pd.Index(["default"]))
            df[idx] = df[idx].apply(lambda x: x.fillna(x["default"]), axis=1)
        return df

    @Slot(name="resetDataIndex")
    def rebuild_table(self) -> None:
        """Should be called when the `parameters_changed` signal is emitted.
        Will call sync with a copy of the current dataframe to ensure no
        user-imported data is lost.

        TODO: handle database parameter group changes correctly. Maybe a
         separate signal like rename?
        """
        self.sync(self._dataframe.reset_index())

    @Slot(str, str, str, name="renameParameterIndex")
    def update_param_name(self, old: str, group: str, new: str) -> None:
        """Kind of a cheat, but directly edit the dataframe.index to update
        the table whenever the user renames a parameter.
        """
        new_idx = pd.Index(
            np.where(
                (self._dataframe.index == old) & (self._dataframe["Group"] == group),
                new,
                self._dataframe.index,
            ),
            name=self._dataframe.index.name,
        )
        self._dataframe.index = new_idx
        self.updated.emit()

    def iterate_scenarios(self) -> Iterable[Tuple[str, Iterable]]:
        """Iterates through all of the non-description columns from left to right.

        Returns an iterator of tuples containing the scenario name and a dictionary
        of the parameter names and new amounts.

        TODO: Fix this so it returns the least amount of required information.
        """
        df = self._dataframe.reset_index()
        df = df.set_index(["Group", "Name"])


        return (
            (scenario, df[scenario])
            for scenario in df.columns
        )



# -*- coding: utf-8 -*-
from importlib import metadata
from os import environ
from conda import cli
import requests


def check_ab_update() -> bool:
    ab_url = "https://api.anaconda.org/package/mrvisscher/activity-browser-beta"
    ab_response = requests.get(ab_url)
    ab_current = metadata.version("activity_browser")

    if ab_response.status_code != 200:
        print("Could not fetch latest activity browser beta version")
        return False

    ab_latest = ab_response.json()['latest_version'].replace(".", "")

    print(f"activity_browser_beta: {ab_current} x {ab_latest}")

    if ab_current == "0.0.0" or ab_current == ab_latest:
        return False
    return True


def run():
    from activity_browser import run_activity_browser
    print("Launching the Activity Browser")
    run_activity_browser()


def run_activity_browser():
    print("Activity Browser 3 Beta Release")
    print("______________________________________")
    ab = check_ab_update()
    print("______________________________________")
    if ab and environ.get("CONDA_DEFAULT_ENV"):
        print("Updating activity-browser-beta")
        cli.main("update", "-c", "mrvisscher", "activity-browser-beta",)
    run()



from qtpy import QtWidgets
from typing import Callable, Optional
from inspect import signature


class ABMenu(QtWidgets.QMenu):
    menuSetup: list[Callable[["ABMenu", Optional[QtWidgets.QWidget]], None]]
    title: str = None

    def __init__(self, pos, parent=None):
        super().__init__(parent)

        for item in self.menuSetup:
            if len(signature(item).parameters) == 1:
                item(self)
            if len(signature(item).parameters) == 2:
                item(self, parent)

    def add(self, action, *args, enable=True, text=None, **kwargs):
        qaction = action.get_QAction(*args, parent=self, enabled=enable, text=text, **kwargs)
        self.addAction(qaction)



from typing import Any, List

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ExchangeDelete(ABAction):
    """
    ABAction to delete one or more exchanges from an activity.
    """

    icon = qicons.delete
    text = "Delete exchange(s)"

    @staticmethod
    @exception_dialogs
    def run(exchanges: List[Any]):
        for exchange in exchanges:
            exchange.delete()



# -*- coding: utf-8 -*-
import os

from qtpy import QtWidgets
from qtpy.QtCore import Slot

from activity_browser.bwutils import exporters as exp
from activity_browser.mod import bw2data as bd

EXPORTERS = {
    # Store data as a BW2Package.
    "BW2Package": exp.store_database_as_package,
    # Export the database, all project parameters and all parameters that are
    # related to that database as an Excel file.
    "Excel": exp.write_lci_excel,
}
EXTENSIONS = {
    "BW2Package": ".bw2package",
    "Excel": ".xlsx",
}


class DatabaseExportWizard(QtWidgets.QWizard):
    """Present the user with a wizard that assist in either importing
    or exporting a database.
    On either import or exporting, present a progress bar
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle("Database export wizard")
        self.export_page = ExportDatabasePage(self)
        self.pages = [self.export_page]
        for i, page in enumerate(self.pages):
            self.setPage(i, page)

    def accept(self) -> None:
        self.perform_export()
        super().accept()

    def perform_export(self) -> None:
        db_name = self.field("database_choice")
        export_as = self.field("export_option")
        out_path = self.field("output_path")
        # Ensure that extension matches export_option.
        path, ext = os.path.splitext(out_path)
        if ext and not ext == EXTENSIONS[export_as]:
            ext = EXTENSIONS[export_as]
            out_path = path + ext
        elif not ext:
            out_path = path + EXTENSIONS[export_as]
        EXPORTERS[export_as](db_name, out_path)


class ExportDatabasePage(QtWidgets.QWizardPage):
    FILTERS = {
        "BW2Package": "BW2Package Files (*.bw2package);; All Files (*.*)",
        "Excel": "Excel Files (*.xlsx);; All Files (*.*)",
    }

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.wizard = parent
        self.database = QtWidgets.QComboBox()
        self.export_option = QtWidgets.QComboBox()
        self.export_option.addItems(list(EXPORTERS))
        self.database.currentIndexChanged.connect(self.changed)
        self.output_dir = QtWidgets.QLineEdit()
        self.output_dir.setReadOnly(True)
        self.browse_button = QtWidgets.QPushButton("Browse")
        self.browse_button.clicked.connect(self.browse)
        self.complete = False

        box = QtWidgets.QGroupBox("Database selection:")
        grid = QtWidgets.QGridLayout()
        grid.addWidget(QtWidgets.QLabel("Database:"), 0, 0, 1, 1)
        grid.addWidget(self.database, 0, 1, 1, 2)
        grid.addWidget(QtWidgets.QLabel("Exported as:"), 1, 0, 1, 1)
        grid.addWidget(self.export_option, 1, 1, 1, 2)
        grid.addWidget(
            QtWidgets.QLabel("Exported data is stored in the directory below:"),
            2,
            0,
            1,
            3,
        )
        grid.addWidget(self.output_dir, 3, 0, 1, 2)
        grid.addWidget(self.browse_button, 3, 2, 1, 1)
        box.setLayout(grid)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(box)
        self.setLayout(layout)

        self.setFinalPage(True)
        self.registerField("database_choice", self.database, "currentText")
        self.registerField("export_option", self.export_option, "currentText")
        self.registerField("output_path*", self.output_dir)

    def initializePage(self):
        self.wizard.setButtonLayout(
            [
                QtWidgets.QWizard.Stretch,
                QtWidgets.QWizard.FinishButton,
                QtWidgets.QWizard.CancelButton,
            ]
        )
        self.database.clear()
        choices = ["-----"] + list(bd.databases)
        self.database.addItems(choices)
        self.output_dir.setText(str(bd.projects.output_dir))

    def changed(self):
        self.complete = False if self.database.currentText() == "-----" else True
        self.completeChanged.emit()

    def isComplete(self):
        return self.complete

    @Slot(name="browseFile")
    def browse(self) -> None:
        file_filter = self.FILTERS[self.field("export_option")]
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent=self, caption="Save database", filter=file_filter
        )
        if path:
            self.output_dir.setText(path)



from logging import getLogger

from bw2io import *

import bw2io.remote as remote




log = getLogger(__name__)


def ab_bw2setup(version):
    import bw2io as bi
    from activity_browser.mod.bw2io.importers.ecospold2_biosphere import ABEcospold2BiosphereImporter
    from activity_browser.info import __ei_versions__
    from activity_browser.utils import sort_semantic_versions
    from .migrations import ab_create_core_migrations

    ab_create_core_migrations()

    version = version[:3]

    if version == sort_semantic_versions(__ei_versions__)[0][:3]:
        log.info(f"Installing biosphere version >{version}<")
        # most recent version
        bio_import = ABEcospold2BiosphereImporter()
    else:
        log.info(f"Installing legacy biosphere version >{version}<")
        # not most recent version, import legacy biosphere from AB
        bio_import = ABEcospold2BiosphereImporter(version=version)
    bio_import.apply_strategies()
    log.info("Writing biosphere database")
    bio_import.write_database()

    log.info("Writing LCIA methods")
    create_default_lcia_methods()

    # patching biosphere
    sorted_versions = sort_semantic_versions(
        __ei_versions__, highest_to_lowest=False
    )
    ei_versions = sorted_versions[: sorted_versions.index(version) + 1]

    patches = [
        patch
        for patch in dir(bi.data)
        if patch.startswith("add_ecoinvent")
        and patch.endswith("biosphere_flows")
        and any(version.replace(".", "") in patch for version in ei_versions)
    ]

    for patch in patches:
        log.info(f"Applying biosphere patch: {patch}")
        update_bio = getattr(bi.data, patch)
        update_bio()




from qtpy import QtWidgets, QtCore

from activity_browser import application, settings, signals
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

from .database_open import DatabaseOpen


class DatabaseNew(ABAction):
    """
    Executes the process of creating a new database.

    This method retrieves the database name and backend type from the `NewDatabaseDialog`.
    It validates the input to ensure the name is not empty and does not already exist.
    If the input is valid, it creates and registers a new database, then opens it.

    Steps:
    - Open the `NewDatabaseDialog` to get the database name and backend type.
    - Return if the dialog is canceled or the name is empty.
    - Check if the database name already exists and show an error message if it does.
    - Create a new database with the specified name and backend type.
    - Register the database with default settings (not searchable, not read-only).
    - Open the newly created database.

    Raises:
        None
    """
    icon = qicons.add
    text = "New database..."
    tool_tip = "Make a new database"

    @staticmethod
    @exception_dialogs
    def run():
        name, backend, ok = NewDatabaseDialog.get_new_database_data()

        if not ok or not name:
            return

        if name in bd.databases:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible",
                "A database with this name already exists.",
            )
            return

        db = bd.Database(name, backend if backend else "functional_sqlite")
        db.register(searchable=False, read_only=False)

        DatabaseOpen.run([name])


class NewDatabaseDialog(QtWidgets.QDialog):
    """
    A dialog for creating a new database.
    """

    def __init__(self, window_title="New Database", backend="functional_sqlite", parent=None):
        super().__init__(parent)
        self.setWindowTitle(window_title)
        self.setModal(True)

        self.name_input = QtWidgets.QLineEdit(self)
        self.name_input.setPlaceholderText("Enter database name")
        self.name_input.textChanged.connect(self.validate)

        self.backend_dropdown = QtWidgets.QComboBox(self)
        self.backend_dropdown.addItems(["functional_sqlite", "sqlite"])
        self.backend_dropdown.setCurrentText(backend)

        self.create_button = QtWidgets.QPushButton("Create", self)
        self.create_button.setDisabled(True)
        self.create_button.clicked.connect(self.accept)

        self.build_layout()

    @classmethod
    def get_new_database_data(cls, window_title="New Database", backend="functional_sqlite") -> tuple[str, str, bool]:
        """
        Opens a dialog to collect data for creating a new database.

        Returns:
            tuple[str, str, bool]: A tuple containing:
                - The name of the new database (str).
                - The selected backend type (str).
                - A boolean indicating whether the dialog was accepted (True) or canceled (False).
        """
        dialog = cls(window_title, backend, application.main_window)
        result = dialog.exec_()

        return dialog.name_input.text(), dialog.backend_dropdown.currentText(), result == QtWidgets.QDialog.Accepted


    def build_layout(self):
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.name_input)
        layout.addWidget(QtWidgets.QLabel("Select backend:", self))
        layout.addWidget(self.backend_dropdown)
        layout.addWidget(self.create_button, alignment=QtCore.Qt.AlignRight)
        self.setLayout(layout)

    def validate(self, text):
        if text in bd.databases or not text:
            self.create_button.setDisabled(True)
        else:
            self.create_button.setDisabled(False)





import hashlib
import textwrap
from datetime import datetime
from logging import getLogger
from collections import OrderedDict

import arrow
import pandas as pd
import peewee as pw

import bw2data as bd
from bw2data.parameters import ParameterBase, ProjectParameter, DatabaseParameter, ActivityParameter, Group
from bw2data.errors import UnknownObject

from functools import lru_cache

from .metadata import AB_metadata
from .utils import Parameter

log = getLogger(__name__)

"""
bwutils is a collection of methods that build upon brightway2 and are generic enough to provide here so that we avoid
re-typing the same code in different parts of the Activity Browser.

When adding new methods, please use the sections below (or add a new section, if required).
"""


# Formatting
def wrap_text(string: str, max_length: int = 80) -> str:
    """Wrap the label making sure that key and name are in 2 rows.

    idea from https://stackoverflow.com/a/39134215/4929813
    """

    def fold(line: str) -> str:
        return textwrap.fill(
            line, width=max_length, break_long_words=True, replace_whitespace=False
        )

    return "\n".join(map(fold, string.splitlines()))


def format_activity_label(key, style="pnl", max_length=40):
    try:
        act = bd.get_activity(key)

        if style == "pnl":
            label = "\n".join(
                [
                    act.get("reference product", ""),
                    act.get("name", ""),
                    str(act.get("location", "")),
                ]
            )
        elif style == "pnl_":
            label = " | ".join(
                [
                    act.get("reference product", ""),
                    act.get("name", ""),
                    str(act.get("location", "")),
                ]
            )
        elif style == "pnld":
            label = " | ".join(
                [
                    act.get("reference product", ""),
                    act.get("name", ""),
                    str(act.get("location", "")),
                    act.get("database", ""),
                ]
            )
        elif style == "pl":
            label = ", ".join(
                [
                    act.get("reference product", "") or act.get("name", ""),
                    str(act.get("location", "")),
                ]
            )
        elif style == "key":
            label = str(act.key)  # safer to use key, code does not always exist

        elif style == "bio":
            label = ",\n".join([act.get("name", ""), str(act.get("categories", ""))])
        else:
            label = "\n".join(
                [
                    act.get("reference product", ""),
                    act.get("name", ""),
                    str(act.get("location", "")),
                ]
            )
    except:
        if isinstance(key, tuple):
            return wrap_text(str("".join(key)))
        else:
            return wrap_text(str(key))
    return wrap_text(label, max_length=max_length)


def cleanup_deleted_bw_projects() -> None:
    """Clean up the deleted projects from disk.

    NOTE: This cannot be done from within the AB.
    """
    n_dir = bd.projects.purge_deleted_directories()
    log.info(f"Deleted {n_dir} unused project directories!")


def projects_by_last_opened():
    def key(ds):
        if not ds.data or "last_opened" not in ds.data:
            return 0
        date = datetime.fromisoformat(ds.data["last_opened"])

        return int(date.strftime("%Y%m%d%H%M%S"))

    projects = list(bd.projects)
    projects.sort(key=key, reverse=True)

    return projects



# Database
def get_database_metadata(name):
    """Returns a dictionary with database meta-information."""
    d = dict()
    d["Name"] = name
    d["Depends"] = "; ".join(bd.databases[name].get("depends", []))
    dt = bd.databases[name].get("modified", "")
    if dt:
        dt = arrow.get(dt).humanize()
    d["Last modified"] = dt
    return d

def database_is_locked(name: str) -> bool:
    """Returns True if the database is locked, False otherwise."""
    if not name in bd.databases:
        raise KeyError("Not an existing database:", name)
    return bd.databases[name].get("read_only", True)

def database_is_legacy(name: str) -> bool:
    """Returns True if the database is locked, False otherwise."""
    if not name in bd.databases:
        raise KeyError("Not an existing database:", name)

    database = bd.Database(name)

    if database.backend not in ["sqlite", "functional_sqlite"]:
        raise ValueError("Database backend must be sqlite or functional_sqlite")

    return database.backend == "sqlite"

def is_technosphere_db(db_name: str) -> bool:
    """Returns True if database describes the technosphere, False if it describes a biosphere."""
    if not db_name in bd.databases:
        raise KeyError("Not an existing database:", db_name)
    # This code seems incorrect, just return True for now
    return True


def count_database_records(name: str) -> int:
    """To account for possible brightway database types that do not implement
    the __len__ method.
    """
    try:
        return len(AB_metadata.dataframe.loc[name])
    except KeyError:
        return 0



# Activity
AB_names_to_bw_keys = {
    "Amount": "amount",
    "Product": "reference product",
    "Activity": "name",
    "Unit": "unit",
    "Location": "location",
    "Database": "database",
    "Uncertainty": "uncertainty type",
    "Formula": "formula",
    "Categories": "categories",
    "Type": "type",
    "Comment": "comment",
    "Functional": "functional",
}

bw_keys_to_AB_names = {v: k for k, v in AB_names_to_bw_keys.items()}


def get_activity_name(key, str_length=22):
    return ",".join(key.get("name", "").split(",")[:3])[:str_length]


def refresh_node(node: tuple | int | bd.Node) -> bd.Node:
    if isinstance(node, bd.Node):
        node = bd.get_node(id=node.id)
    elif isinstance(node, tuple):
        node = bd.get_node(key=node)
    elif isinstance(node, int):
        node = bd.get_node(id=node)
    else:
        raise ValueError("Activity must be either a tuple, int or Node instance")
    return node


def refresh_node_or_none(node: tuple | int | bd.Node) -> bd.Node | None:
    try:
        return refresh_node(node)
    except (ValueError, UnknownObject):
        return None


def refresh_parameter(parameter: tuple | Parameter | ParameterBase):
    # construct mock-parameter if it is a parameter key (group, name)
    if isinstance(parameter, tuple) and len(parameter) == 2:
        if parameter[0] == "project":
            parameter = Parameter(parameter[1], parameter[0], None, None, "project")
        elif parameter[0] in bd.databases:
            parameter = Parameter(parameter[1], parameter[0], None, None, "database")
        else:
            parameter = Parameter(parameter[1], parameter[0], None, None, "activity")

    # get the newest peewee model from the database
    if isinstance(parameter, Parameter):
        raw = parameter.to_peewee_model()
    elif isinstance(parameter, ParameterBase):
        raw = parameter.get_by_id(parameter.id)
    else:
        raise ValueError("Unknown parameter type")

    # construct a new Parameter-type from the peewee model
    if isinstance(raw, ProjectParameter):
        return Parameter(raw.dict.pop("name"), "project", raw.dict.get("amount"), raw.dict, "project")
    elif isinstance(raw, DatabaseParameter):
        return Parameter(raw.dict.pop("name"), raw.database, raw.dict.get("amount"), raw.dict, "database")
    elif isinstance(raw, ActivityParameter):
        return Parameter(raw.dict.pop("name"), raw.group, raw.dict.get("amount"), raw.dict, "activity")
    else:
        raise ValueError("Unknown parameter type")


def parameters_in_scope(
        node: tuple | int | bd.Node = None,
        parameter: tuple | Parameter | ParameterBase = None
) -> dict[str, Parameter]:
    if (not node and not parameter) or (node and parameter):
        raise ValueError("Supply either node or parameter")
    if node:
        node = refresh_node(node)
        database = node["database"]
        group = node_group(node)
    else:  # if parameter
        parameter = refresh_parameter(parameter)
        group = parameter.group
        if group == "project":
            database = None
        elif group in bd.databases:
            database = group
        else:
            database = ActivityParameter.get_or_none(group=group).database

    data = OrderedDict()

    for name, param in ProjectParameter.load().items():
        data[name] = Parameter(name, "project", param["amount"], param, "project")

    for name, param in DatabaseParameter.load(database).items():
        if name in data:
            del data[name]  # the variable is overwritten in the scope chain
        data[name] = Parameter(name, database, param["amount"], param, "database")

    group_deps = Group.get_or_none(name=group).order + [group] if group else []

    for dep in group_deps:
        for name, param in ActivityParameter.load(dep).items():
            if name in data:
                del data[name]  # the variable is overwritten in the scope chain
            data[name] = Parameter(name, dep, param.get("amount"), param, "activity")

    return data


def node_group(node: tuple | int | bd.Node) -> str | None:
    """Returns the group of the node, or None if it does not have a group."""
    node = refresh_node(node)
    ap = ActivityParameter.get_or_none(database=node["database"], code=node["code"])
    return ap.group if ap else None


def clean_activity_name(activity_name: str) -> str:
    """Takes a given activity name and remove or replace all characters
    not allowed to be in there.

    Use this when creating parameters, as there are specific characters not
    allowed to be in parameter names.

    These are ' -,.%[]'
    Integers are also removed aggressively, there are allowed, but not
    at the start of a parameter name.
    """
    remove = ",.%[]()0123456789"
    replace = " -"
    # Remove invalid characters
    for char in remove:
        if char in activity_name:
            activity_name = activity_name.replace(char, "")
    # Replace spacing and dashes with underscores
    for char in replace:
        if char in activity_name:
            activity_name = activity_name.replace(char, "_")
    # strip underscores from start of string
    activity_name = activity_name.lstrip("_")
    return activity_name


def build_activity_group_name(key: tuple, name: str = None) -> str:
    """Constructs a group name unique to a given bw activity.

    If given a `name`, use that instead of looking up the activity name.

    NOTE: The created group name is not easy for users to understand, so hide
    it from them where possible.
    """
    simple_hash = hashlib.md5(":".join(key).encode()).hexdigest()
    if name:
        return "{}_{}".format(name, simple_hash)
    act = bd.get_activity(key)
    clean = clean_activity_name(act.get("name"))
    return "{}_{}".format(clean, simple_hash)


@lru_cache(maxsize=2048)
def identify_activity_type(activity):
    """Return the activity type based on its naming."""
    name = activity["name"]
    if "treatment of" in name:
        return "treatment"
    elif "market for" in name:
        # if not "to generic" in name:  # these are not markets, but also transferring activities
        return "market"
    elif "market group" in name:
        # if not "to generic" in name:
        return "marketgroup"
    else:
        return "production"


def generate_copy_code(key: tuple) -> str:
    """Generate a new code to use when copying an activity"""
    db, code = key
    metadata = AB_metadata.get_database_metadata(db)
    if "_copy" in code:
        code = code.split("_copy")[0]
    copies = (
        metadata["key"]
        .apply(lambda x: x[1] if code in x[1] and "_copy" in x[1] else None)
        .dropna()
        .to_list()
        if not metadata.empty
        else []
    )
    if not copies:
        return f"{code}_copy1"
    n = max((int(c.split("_copy")[1]) for c in copies))
    return f"{code}_copy{n + 1}"


# EXCHANGES
def refresh_edge(edge: int | bd.Edge) -> bd.Edge:
    if isinstance(edge, bd.Edge):
        edge = edge.__class__(bd.Edge.ORMDataset.get_by_id(edge.id))
    elif isinstance(edge, int):
        # Edge is just the ID
        # need to go at it using a workaround to get the edge class right through the owner activity
        owner = bd.Edge(bd.Edge.ORMDataset.get_by_id(edge)).output
        for candidate in owner.edges():
            if candidate.id == edge:
                return candidate
    else:
        raise ValueError("Edge must be either an int or Edge instance")
    return edge


def refresh_edge_or_none(edge: int | bd.Edge) -> bd.Edge | None:
    try:
        return refresh_edge(edge)
    except (ValueError, UnknownObject):
        return None


def get_exchanges_in_scenario_difference_file_notation(exchanges):
    """From a list of exchanges get the information needed for the scenario difference (SDF) file that is used in
    conjunction with the superstructure approach. This is a convenience function to export data from the AB in a format
    suitable for the SDF."""
    data = []
    for exc in exchanges:
        try:
            from_act = bd.get_activity(exc.get("input"))
            to_act = bd.get_activity(exc.get("output"))

            row = {
                "from activity name": from_act.get("name", ""),
                "from reference product": from_act.get("reference product", ""),
                "from location": from_act.get("location", ""),
                "from categories": from_act.get("categories", ""),
                "from database": from_act.get("database", ""),
                "from key": from_act.key,
                "to activity name": to_act.get("name", ""),
                "to reference product": to_act.get("reference product", ""),
                "to location": to_act.get("location", ""),
                "to categories": to_act.get("categories", ""),
                "to database": to_act.get("database", ""),
                "to key": to_act.key,
                "flow type": exc.get("type", ""),
                "amount": exc.get("amount", ""),
            }
            data.append(row)

        except:
            # The input activity does not exist. remove the exchange.
            log.error(
                "Something did not work with the following exchange: {}. It was removed from the list.".format(
                    exc
                )
            )
    return data


def exchanges_to_sdf(exchanges: list[dict]) -> pd.DataFrame:
    data = get_exchanges_in_scenario_difference_file_notation(exchanges)
    return pd.DataFrame(data)


def get_exchanges_from_a_list_of_activities(
    activities: list, as_keys: bool = False
) -> list:
    """Get all exchanges in a list of activities."""
    if as_keys:
        activities = [bd.get_activity(key) for key in activities]
    exchanges = []
    for act in activities:
        for exc in act.exchanges():
            exchanges.append(exc)
    return exchanges


# LCIA
def unit_of_method(method: tuple) -> str:
    """Attempt to return the unit of the given method."""
    assert method in bd.methods
    return bd.methods[method].get("unit", "unit")


def get_LCIA_method_name_dict(keys: list) -> dict:
    """Impact categories in brightway2 are stored in tuples, which is
    unpractical for display in, e.g. dropdown menus.

    Returns a dictionary with
        keys: comma separated strings
        values: brightway2 method tuples
    """
    return {", ".join(key): key for key in keys}



from pyprind.progbar import *
from activity_browser.mod.patching import patch_superclass, patched
from activity_browser.ui.threading import thread_local

from qtpy.QtCore import QObject, SignalInstance, Signal

@patch_superclass
class Progbar(ProgBar):

    def _print(self, force_flush=False):
        patched[Progbar]["_print"](self, force_flush)
        try:
            thread_local.progress_slot(int(self._calc_percent()), self.title)
            qt_pyprind.updated.emit(self.title, self._calc_percent())
        except AttributeError:
            pass


class QtPyprind(QObject):
    updated: SignalInstance = Signal(str, float)


qt_pyprind = QtPyprind()




from qtpy import QtWidgets, QtGui

import bw2data as bd

from activity_browser import bwutils, actions


class DescriptionTab(QtWidgets.QTextEdit):
    """
    A widget that displays and edits the description (comment) of a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display and edit the description for.
    """
    def __init__(self, activity: tuple | int | bd.Node, parent=None):
        """
        Initializes the DescriptionTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display and edit the description for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        self.activity = bwutils.refresh_node(activity)
        super().__init__(parent, self.activity.get("comment", ""))
        self.setPlaceholderText("Click here to edit the description of this activity...")

    def sync(self):
        """
        Synchronizes the widget with the current state of the activity.
        """
        self.activity = bwutils.refresh_node(self.activity)
        self.setText(self.activity.get("comment", ""))
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)

        # Set the read-only state based on the activity's database
        self.setDisabled(bwutils.database_is_locked(self.activity["database"]))

    def focusOutEvent(self, e):
        """
        Handles the focus out event to save the comment if it has changed.

        Args:
            e: The focus out event.
        """
        if self.toPlainText() == self.activity.get("comment", ""):
            return
        actions.ActivityModify.run(self.activity, "comment", self.toPlainText())



from tqdm import tqdm
from logging import getLogger
from qtpy import QtWidgets, QtGui, QtCore

import bw2data as bd
import pandas as pd

from activity_browser import application, signals
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import AB_metadata
from activity_browser.ui.icons import qicons
from activity_browser.ui.threading import ABThread

log = getLogger(__name__)


class ProjectMigrate25(ABAction):
    """
    ABAction to duplicate a project. Asks the user for a new name. Returns if no name is given, the user cancels, or
    when the name is already in use by another project. Else, instructs the ProjectController to duplicate the current
    project to the new name.
    """

    icon = qicons.copy
    text = "Migrate project"
    tool_tip = "Migrate the project to bw25"

    @staticmethod
    @exception_dialogs
    def run(name: str = None):
        if name is None:
            name = bd.projects.current

        dialog = MigrateDialog(name, application.main_window)
        dialog.exec_()

        if dialog.result() == dialog.DialogCode.Rejected:
            return

        if name != bd.projects.current:
            bd.projects.set_current(name, update=False)

        # setup dialog
        progress = QtWidgets.QProgressDialog(
            parent=application.main_window,
            labelText="Migrating project, this may take a while...",
            maximum=0
        )
        progress.setCancelButton(None)
        progress.setWindowTitle("Migrating project to Brightway25")
        progress.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)
        progress.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        progress.findChild(QtWidgets.QProgressBar).setTextVisible(False)
        progress.resize(400, 100)
        progress.show()

        thread = MigrateThread(application)
        thread.finished.connect(lambda: progress.deleteLater())
        thread.start()
        thread.connect_progress_dialog(progress)


class MigrateDialog(QtWidgets.QDialog):
    def __init__(self, project_name: str, parent=None):
        from .project_export import ProjectExport

        super().__init__(parent)
        self.setWindowTitle("Migrate project")
        label = QtWidgets.QLabel(f"Migrate project ({project_name}) from legacy to Brightway25. This cannot be undone.")

        cancel = QtWidgets.QPushButton("Cancel")
        migrate = QtWidgets.QPushButton("Migrate")
        backup = ProjectExport.get_QButton(project_name)
        backup.setText("Back-up project")
        backup.setIcon(QtGui.QIcon())

        cancel.clicked.connect(self.reject)
        migrate.clicked.connect(self.accept)

        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addWidget(backup)
        button_layout.addStretch()
        button_layout.addWidget(cancel)
        button_layout.addWidget(migrate)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(label)
        layout.addLayout(button_layout)

        self.setLayout(layout)


class MigrateThread(ABThread):
    def run_safely(self):
        self.pre_process_methods()

        log.info("Updating and processing all datasets in the project")
        bd.projects.set_current(bd.projects.current)

        for db_name in bd.databases:
            self.update_database_activity_types(db_name)

        # set the bw25 flag in the project dataset
        bd.projects.dataset.data["25"] = True
        bd.projects.dataset.save()

        # reloading project to ensure all changes are applied
        bd.projects.set_current(bd.projects.current)

    @classmethod
    def pre_process_methods(cls):
        log.info("Pre-processing methods for migration to bw25")
        data = {m: bd.Method(m).load() for m in bd.methods}
        df = pd.DataFrame([(k, v[0][0], v[0][1], v[1])
                           for k, values in data.items() for v in values
                           if isinstance(v[0], tuple) and len(v) == 2 and len(v[0]) == 2],
                          columns=["method", "database", "code", "value"])

        df = df.merge(AB_metadata.dataframe["id"], left_on=["database", "code"], right_index=True)

        signals.method.blockSignals(True)
        signals.meta.blockSignals(True)

        for name in tqdm(df["method"].unique(), desc="Pre-processing methods", unit="method", total=len(df["method"].unique())):
            method_df = df[df["method"] == name][["id", "value"]]
            method_list = list(method_df.itertuples(index=False, name=None))
            bd.Method(name).write(method_list, process=False)

        signals.method.blockSignals(False)
        signals.meta.blockSignals(False)

        return

    @classmethod
    def update_database_activity_types(cls, db_name: str):
        database = bd.Database(db_name)
        write = False

        if not isinstance(database, bd.backends.SQLiteBackend):
            return

        log.info(f"Updating activity types in {db_name}")
        raw = database.load()

        for key, ds in tqdm(raw.items(), desc=f"Updating activity types in {db_name}", unit="activity", total=len(raw)):
            if cls.activity_is_processwithreferenceproduct(ds):
                write = True
                ds["type"] = "processwithreferenceproduct"

        if write:
            database.write(raw)

    @staticmethod
    def activity_is_processwithreferenceproduct(ds: dict) -> bool:
        production = [exc for exc in ds.get("exchanges", []) if exc.get("type") == "production"]
        return (
            ds.get("type") in ["process", "processwithreferenceproduct"] and
            (
                len(production) == 0 or
                production[0].get("input") == (ds["database"], ds["code"])
            )
        )






# -*- coding: utf-8 -*-
"""Classes related to the cutoff options menu in contributions tabs.

These classes contain all menu items required to modify the cutoffs of the MLCA results. The
CutoffMenu class is responsible for assembling the menu. Each different menu item is contained in
its separate class.
"""

from collections import namedtuple
from typing import Union

import numpy as np
from qtpy import QtCore
from qtpy.QtCore import QLocale, Qt, Signal, Slot
from qtpy.QtGui import QDoubleValidator, QIntValidator
from qtpy.QtWidgets import (QButtonGroup, QHBoxLayout, QLabel, QLineEdit,
                               QPushButton, QRadioButton, QSlider, QVBoxLayout,
                               QWidget)

from activity_browser.ui import widgets


# These tuples are used in referring to the two Types and three Labels used
Types = namedtuple("types", ("percent", "cum_percent", "number"))
Labels = namedtuple("labels", ("unit", "min", "max"))


class CutoffMenu(QWidget):
    """This class assembles the cutoff menu from the other classes in this module."""

    slider_change = Signal()

    def __init__(self, parent=None, cutoff_value=0.05, limit_type="percent"):
        super().__init__(parent)
        self.cutoff_value = cutoff_value
        self.limit_type = limit_type

        locale = QLocale(QLocale.English, QLocale.UnitedStates)
        locale.setNumberOptions(QLocale.RejectGroupSeparator)
        self.validators = Types(
            QDoubleValidator(0.001, 100.0, 1, self),
            QIntValidator(0, 100, self),
            QIntValidator(0, 50, self),
        )
        self.validators.percent.setLocale(locale)
        self.validators.number.setLocale(locale)
        self.buttons = Types(
            QRadioButton("Minimum %"),
            QRadioButton("Cumulative %"),
            QRadioButton("Top #"))
        self.buttons.percent.setChecked(True)
        self.buttons.percent.setToolTip(
            "This cut-off type shows contributions of at least some percentage "
            "(for example contributions of at least 5% of the total impact)"
        )
        self.buttons.cum_percent.setToolTip(
            "This cut-off type shows contributions that together are some percentage "
            "(for example all highest contributors that together count up to 80% of the total impact)"
        )
        self.buttons.number.setToolTip(
            "This cut-off type shows this number of largest contributors "
            "(for example the top 5 largest contributors)"
        )
        self.button_group = QButtonGroup()
        self.button_group.addButton(self.buttons.percent, 0)
        self.button_group.addButton(self.buttons.cum_percent, 1)
        self.button_group.addButton(self.buttons.number, 2)
        self.button_id_limit_type = {
            0: "percent",
            1: "cum_percent",
            2: "number",
        }
        self.button_group.setExclusive(True)
        self.sliders = Types(
            LogarithmicSlider(self),
            QSlider(Qt.Horizontal, self),
            QSlider(Qt.Horizontal, self))
        self.sliders.percent.setToolTip(
            "This slider sets the cut-off percentage to show"
        )
        self.sliders.cum_percent.setToolTip(
            "This slider sets the cumulative cut-off percentage to show"
        )
        self.sliders.number.setToolTip(
            "This slider sets the amount of highest contributors to show"
        )
        self.units = Types("minimum %", "cumulative %", "number")
        self.labels = Labels(QLabel(), QLabel(), QLabel())
        self.cutoff_slider_line = QLineEdit()
        self.cutoff_slider_line.setToolTip(
            "This entry sets the cut-off amount"
        )
        self.cutoff_slider_line.setLocale(locale)
        self.cutoff_slider_lft_btn = QPushButton("<")
        self.cutoff_slider_lft_btn.setToolTip(
            "This button moves the cut-off value one increment"
        )
        self.cutoff_slider_rght_btn = QPushButton(">")
        self.cutoff_slider_rght_btn.setToolTip(
            "This button moves the cut-off value one increment"
        )

        self.debounce_slider = QtCore.QTimer()
        self.debounce_slider.setInterval(300)
        self.debounce_slider.setSingleShot(True)

        self.debounce_text = QtCore.QTimer()
        self.debounce_text.setInterval(300)
        self.debounce_text.setSingleShot(True)

        self.make_layout()
        self.connect_signals()

    def connect_signals(self):
        """Connect the signals of the menu."""
        # Cut-off types
        self.button_group.buttonClicked.connect(self.cutoff_type_check)
        self.cutoff_slider_lft_btn.clicked.connect(self.cutoff_increment_left_check)
        self.cutoff_slider_rght_btn.clicked.connect(self.cutoff_increment_right_check)

        self.debounce_slider.timeout.connect(self.initiate_slider_change)
        self.debounce_text.timeout.connect(self.initiate_text_change)

        self.sliders.percent.valueChanged.connect(self.debounce_slider.start)
        self.sliders.cum_percent.valueChanged.connect(self.debounce_slider.start)
        self.sliders.number.valueChanged.connect(self.debounce_slider.start)
        self.cutoff_slider_line.textChanged.connect(self.debounce_text.start)

    def initiate_slider_change(self):
        if self.limit_type == "percent":
            self.cutoff_slider_percent_check("sl")
        elif self.limit_type == "cum_percent":
            self.cutoff_slider_cum_percent_check("sl")
        elif self.limit_type == "number":
            self.cutoff_slider_number_check("sl")

    def initiate_text_change(self):
        if self.limit_type == "percent":
            self.cutoff_slider_percent_check("le")
        elif self.limit_type == "cum_percent":
            self.cutoff_slider_cum_percent_check("le")
        elif self.limit_type == "number":
            self.cutoff_slider_number_check("le")

    @Slot(name="incrementLeftCheck")
    def cutoff_increment_left_check(self):
        """Move the slider 1 increment to left when left button is clicked."""
        if self.limit_type == "percent":
            num = int(self.sliders.percent.value())
            self.sliders.percent.setValue(num + 1)
        elif self.limit_type == "cum_percent":
            num = int(self.sliders.cum_percent.value())
            self.sliders.cum_percent.setValue(num - 1)
        elif self.limit_type == "number":
            num = int(self.sliders.number.value())
            self.sliders.number.setValue(num - 1)

    @Slot(name="incrementRightCheck")
    def cutoff_increment_right_check(self):
        """Move the slider 1 increment to right when right button is clicked."""
        if self.limit_type == "percent":
            num = int(self.sliders.percent.value())
            self.sliders.percent.setValue(num - 1)
        elif self.limit_type == "cum_percent":
            num = int(self.sliders.cum_percent.value())
            self.sliders.cum_percent.setValue(num + 1)
        elif self.limit_type == "number":
            num = int(self.sliders.number.value())
            self.sliders.number.setValue(num + 1)

    @Slot(name="isClicked")
    def cutoff_type_check(self) -> None:
        """Dependent on cutoff-type, set the right labels.

                Slot connected to the 'Cut-off types', the state of those buttons determines:
                - which sliders are visible
                - the unit shown
                - minimum and maximum
                - limit_type
                """
        # determine which mode is clicked
        clicked_type = self.button_id_limit_type[self.button_group.checkedId()]
        if self.limit_type == clicked_type:
            return  # immediately return if the clicked type was already toggled
        self.limit_type = clicked_type

        # temporarily block signals
        self.sliders.percent.blockSignals(True)
        self.sliders.cum_percent.blockSignals(True)
        self.sliders.number.blockSignals(True)
        self.cutoff_slider_line.blockSignals(True)

        if self.limit_type == "percent":
            self.sliders.percent.setVisible(True)
            self.sliders.cum_percent.setVisible(False)
            self.sliders.number.setVisible(False)

            self.labels.unit.setText(self.units.percent)
            self.labels.min.setText("100%")
            self.labels.max.setText("0.001%")
            self.cutoff_slider_line.setValidator(self.validators.percent)
        if self.limit_type == "cum_percent":
            self.sliders.percent.setVisible(False)
            self.sliders.cum_percent.setVisible(True)
            self.sliders.number.setVisible(False)

            self.labels.unit.setText(self.units.cum_percent)
            self.labels.min.setText("1%")
            self.labels.max.setText("100%")
            self.cutoff_slider_line.setValidator(self.validators.cum_percent)
        elif self.limit_type == "number":
            self.sliders.percent.setVisible(False)
            self.sliders.cum_percent.setVisible(False)
            self.sliders.number.setVisible(True)

            self.labels.unit.setText(self.units.number)
            self.labels.min.setText(str(self.sliders.number.minimum()))
            self.labels.max.setText(str(self.sliders.number.maximum()))
            self.cutoff_slider_line.setValidator(self.validators.number)

        # unblock signals
        self.sliders.percent.blockSignals(False)
        self.sliders.cum_percent.blockSignals(False)
        self.sliders.number.blockSignals(False)
        self.cutoff_slider_line.blockSignals(False)

    @Slot(str, name="sliderPercentCheck")
    def cutoff_slider_percent_check(self, editor: str):
        """If 'Percent' selected, change the plots and tables to reflect the slider/line-edit."""
        if not self.limit_type == "percent":
            return
        cutoff = 0.05

        # If called by slider
        if editor == "sl":
            self.cutoff_slider_line.blockSignals(True)
            cutoff = abs(self.sliders.percent.log_value)
            self.cutoff_slider_line.setText(str(cutoff))
            self.cutoff_slider_line.blockSignals(False)

        # if called by line edit
        elif editor == "le":
            self.sliders.percent.blockSignals(True)
            if self.cutoff_slider_line.text() == "-":
                cutoff = 0.001
                self.cutoff_slider_line.setText("0.001")
            elif self.cutoff_slider_line.text() == "":
                cutoff = 0.001
            else:
                cutoff = abs(float(self.cutoff_slider_line.text()))

            if cutoff > 100:
                cutoff = 100
                self.cutoff_slider_line.setText(str(cutoff))
            self.sliders.percent.log_value = float(cutoff)
            self.sliders.percent.blockSignals(False)

        self.cutoff_value = cutoff / 100
        self.slider_change.emit()

    @Slot(str, name="sliderCumPercentCheck")
    def cutoff_slider_cum_percent_check(self, editor: str):
        """If 'Percent' selected, change the plots and tables to reflect the slider/line-edit."""
        """If 'Number' selected, change the plots and tables to reflect the slider/line-edit."""
        if not self.limit_type == "cum_percent":
            return
        cutoff = 2

        # If called by slider
        if editor == "sl":
            self.cutoff_slider_line.blockSignals(True)
            cutoff = abs(int(self.sliders.cum_percent.value()))
            self.cutoff_slider_line.setText(str(cutoff))
            self.cutoff_slider_line.blockSignals(False)

        # if called by line edit
        elif editor == "le":
            self.sliders.cum_percent.blockSignals(True)
            if self.cutoff_slider_line.text() == "-":
                cutoff = self.sliders.cum_percent.minimum()
                self.cutoff_slider_line.setText(str(self.sliders.cum_percent.minimum()))
            elif self.cutoff_slider_line.text() == "":
                cutoff = self.sliders.cum_percent.minimum()
            else:
                cutoff = abs(int(self.cutoff_slider_line.text()))

            if cutoff > self.sliders.cum_percent.maximum():
                cutoff = self.sliders.cum_percent.maximum()
                self.cutoff_slider_line.setText(str(cutoff))
            self.sliders.cum_percent.setValue(int(cutoff))
            self.sliders.cum_percent.blockSignals(False)

        self.cutoff_value = cutoff / 100
        self.slider_change.emit()

    @Slot(str, name="sliderNumberCheck")
    def cutoff_slider_number_check(self, editor: str):
        """If 'Number' selected, change the plots and tables to reflect the slider/line-edit."""
        if not self.limit_type == "number":
            return
        cutoff = 2

        # If called by slider
        if editor == "sl":
            self.cutoff_slider_line.blockSignals(True)
            cutoff = abs(int(self.sliders.number.value()))
            self.cutoff_slider_line.setText(str(cutoff))
            self.cutoff_slider_line.blockSignals(False)

        # if called by line edit
        elif editor == "le":
            self.sliders.number.blockSignals(True)
            if self.cutoff_slider_line.text() == "-":
                cutoff = self.sliders.number.minimum()
                self.cutoff_slider_line.setText(str(self.sliders.number.minimum()))
            elif self.cutoff_slider_line.text() == "":
                cutoff = self.sliders.number.minimum()
            else:
                cutoff = abs(int(self.cutoff_slider_line.text()))

            if cutoff > self.sliders.number.maximum():
                cutoff = self.sliders.number.maximum()
                self.cutoff_slider_line.setText(str(cutoff))
            self.sliders.number.setValue(int(cutoff))
            self.sliders.number.blockSignals(False)

        self.cutoff_value = int(cutoff)
        self.slider_change.emit()

    def make_layout(self):
        """Assemble the layout of the cutoff menu.

        Construct the layout for the cutoff menu widget. The initial layout is set to 'Percent'.
        """
        layout = QHBoxLayout()

        # Cut-off types
        cutoff_type = QVBoxLayout()
        cutoff_type_label = QLabel("Cut-off type")

        # Cut-off slider
        cutoff_slider = QVBoxLayout()
        cutoff_slider_set = QVBoxLayout()
        cutoff_slider_label = QLabel("Cut-off level")
        self.sliders.percent.log_value = self.cutoff_value
        self.sliders.percent.setInvertedAppearance(True)
        self.sliders.cum_percent.setValue(self.cutoff_value)
        self.sliders.cum_percent.setMinimum(1)
        self.sliders.cum_percent.setMaximum(100)
        self.sliders.number.setValue(self.cutoff_value)
        self.sliders.number.setMinimum(1)
        self.sliders.number.setMaximum(50)
        cutoff_slider_minmax = QHBoxLayout()
        self.labels.min.setText("100%")
        self.labels.max.setText("0.001%")
        self.labels.unit.setText("minimum %")
        cutoff_slider_ledit = QHBoxLayout()
        self.cutoff_slider_line.setValidator(self.validators.percent)
        self.cutoff_slider_lft_btn.setMaximumWidth(15)
        self.cutoff_slider_rght_btn.setMaximumWidth(15)

        # Assemble types
        cutoff_type.addWidget(cutoff_type_label)
        cutoff_type.addWidget(self.buttons.percent)
        cutoff_type.addWidget(self.buttons.cum_percent)
        cutoff_type.addWidget(self.buttons.number)

        # Assemble slider set
        self.sliders.number.setVisible(False)
        self.sliders.cum_percent.setVisible(False)
        cutoff_slider_set.addWidget(cutoff_slider_label)
        cutoff_slider_minmax.addWidget(self.labels.min)
        cutoff_slider_minmax.addWidget(self.sliders.percent)
        cutoff_slider_minmax.addWidget(self.sliders.cum_percent)
        cutoff_slider_minmax.addWidget(self.sliders.number)
        cutoff_slider_minmax.addWidget(self.labels.max)
        cutoff_slider_set.addLayout(cutoff_slider_minmax)

        cutoff_slider_ledit.addWidget(self.cutoff_slider_line)
        cutoff_slider_ledit.addWidget(self.cutoff_slider_lft_btn)
        cutoff_slider_ledit.addWidget(self.cutoff_slider_rght_btn)
        cutoff_slider_ledit.addWidget(self.labels.unit)
        cutoff_slider_ledit.addStretch(1)

        cutoff_slider.addLayout(cutoff_slider_set)
        cutoff_slider.addLayout(cutoff_slider_ledit)

        # Assemble cut-off menu
        layout.addLayout(cutoff_type)
        layout.addWidget(widgets.ABVLine(self))
        layout.addLayout(cutoff_slider)
        layout.addStretch()

        self.setLayout(layout)


class LogarithmicSlider(QSlider):
    """Makes a QSlider object that behaves logarithmically.

    Inherits from QSlider. This class uses the property `log_value` getter and setter to modify
    the QSlider through the `value` and `setValue` methods.
    """

    def __init__(self, parent=None):
        super().__init__(Qt.Horizontal, parent)
        self.setMinimum(1)
        self.setMaximum(100)

    @property
    def log_value(self) -> Union[int, float]:
        """Read (slider) and modify from 1-100 to 0.001-100 logarithmically with relevant rounding.

        This function converts the 1-100 values and modifies these to 0.001-100 on a logarithmic
        scale. Rounding is done based on magnitude.
        """

        # Logarithmic math refresher:
        # BOP = Base, Outcome Power;
        # log(B)(O) = P --> log(2)(64) = 6  ||  log(10)(1000) = 3
        #       B^P = O -->        2^6 = 64 ||           10^3 = 1000

        value = float(self.value())
        log_val = np.log10(value)
        power = log_val * 2.5 - 3
        ret_val = np.power(10, power)
        test_val = np.log10(ret_val)

        if test_val < -1:
            return ret_val.round(3)
        elif test_val < -0:
            return ret_val.round(2)
        elif test_val < 1:
            return ret_val.round(1)
        else:
            return ret_val.astype(int)

    @log_value.setter
    def log_value(self, value: float) -> None:
        """Modify value from 0.001-100 to 1-100 logarithmically and set slider to value."""
        value = int(float(value) * np.power(10, 3))
        log_val = np.log10(value).round(3)
        set_val = log_val * 20
        self.setValue(set_val)



from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application, signals
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class CSOpen(ABAction):
    text = "Open"

    @staticmethod
    @exception_dialogs
    def run(cs_names: str | list[str]):
        from activity_browser.layouts import pages

        if isinstance(cs_names, str):
            cs_names = [cs_names]

        for cs_name in cs_names:
            if cs_name not in bd.calculation_setups:
                log.warning(f"Calculation setup {cs_name} not found")
                continue

            page = pages.CalculationSetupPage(cs_name)
            central = application.main_window.centralWidget()

            central.addToGroup("LCA Setup", page)



import shutil

from qtpy import QtWidgets

import bw2data as bd
from bw2data.project import ProjectDataset
from bw2data.utils import safe_filename

from activity_browser import settings, application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons

from .project_switch import ProjectSwitch


class ProjectDelete(ABAction):
    """
    Deletes the specified projects or the currently active project if no project names are provided.

    This method handles the deletion of Brightway2 projects. It ensures that the startup project
    cannot be deleted, prompts the user for confirmation, and optionally deletes the project
    directories from the hard disk.

    Args:
        project_names (list of str, optional): A list of project names to delete. If None, the
                                               currently active project is selected.

    Steps:
    - If no project names are provided, use the currently active project.
    - Return immediately if the project list is empty.
    - Prevent deletion of the startup project and notify the user if attempted.
    - Open a confirmation dialog for the user to approve the deletion.
    - If the user cancels, return without deleting.
    - If the currently active project is being deleted, switch to the startup project.
    - Delete the specified projects, optionally removing their directories from the hard disk.
    - Notify the user of successful deletion.

    Raises:
        None
    """

    icon = qicons.delete
    text = "Delete this project"
    tool_tip = "Delete the project"

    @staticmethod
    @exception_dialogs
    def run(project_names: [str] = None):
        if project_names is None:
            # get the current project
            project_names = [bd.projects.current]

        if len(project_names) == 0:
            return

        # if it's the startup project: reject deletion and inform user
        if settings.ab_settings.startup_project in project_names:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible",
                "Can't delete the startup project. Please select another startup project in the settings first.",
            )
            return

        # open a delete dialog for the user to confirm, return if user rejects
        delete_dialog = ProjectDeletionDialog(project_names, application.main_window)
        if delete_dialog.exec_() != ProjectDeletionDialog.Accepted:
            return

        # try to delete the project, delete directory if user specified so
        if bd.projects.current in project_names:
            ProjectSwitch.run(settings.ab_settings.startup_project)

        for project in project_names:
            ProjectDelete.delete_project(project, delete_dialog.deletion_warning_checked())

        # inform the user of successful deletion
        QtWidgets.QMessageBox.information(
            application.main_window, "Project(s) deleted", "Project(s) successfully deleted"
        )

    @staticmethod
    def delete_project(name: str, delete_dir: bool):

        ds = ProjectDataset.get(ProjectDataset.name == name)

        if delete_dir:
            dir_path = bd.projects._base_data_dir / safe_filename(name, full=ds.full_hash)
            assert dir_path.is_dir(), "Can't find project directory"
            shutil.rmtree(dir_path)

        ds.delete_instance()


class ProjectDeletionDialog(QtWidgets.QDialog):

    def __init__(self, projects: [str], parent=None):
        super().__init__(parent)

        self.title = "Confirm project deletion"

        if len(projects) == 1:
            self.label = QtWidgets.QLabel(
                f"Final confirmation to remove project: {projects[0]}.\n"
                + "Warning: Non reversible process!"
            )
        else:
            self.label = QtWidgets.QLabel(
                f"Final confirmation to remove {len(projects)} projects.\n"
                + "Warning: Non reversible process!"
            )
        self.check = QtWidgets.QVBoxLayout()
        self.hd_check = QtWidgets.QCheckBox(f"Remove from the hard disk")
        self.hd_check.setChecked(True)

        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        self.setWindowTitle(self.title)
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.hd_check)
        self.layout.addWidget(self.buttons)

        self.setLayout(self.layout)

    def deletion_warning_checked(self):
        return self.hd_check.isChecked()
